import time
from datetime import datetime
from collections import defaultdict
import logging

logger = logging.getLogger(__name__)


def calculate_app_stats(app, connection, cursor, fields_filter=None):
    queries = {
        'total_incidents': 'SELECT COUNT(*) FROM `incident` WHERE `created` < %(date_variable)s AND `application_id` = %(application_id)s',
        'total_messages_sent': 'SELECT COUNT(*) FROM `message` WHERE `sent` < %(date_variable)s AND `application_id` = %(application_id)s',
        'total_incidents_last_week': 'SELECT COUNT(*) FROM `incident` WHERE `created` > %(date_variable)s - INTERVAL 1 WEEK AND `created` < %(date_variable)s AND `application_id` = %(application_id)s',
        'total_messages_sent_last_week': 'SELECT COUNT(*) FROM `message` USE INDEX (ix_message_sent) WHERE `sent` > %(date_variable)s - INTERVAL 1 WEEK AND `sent` < %(date_variable)s AND `application_id` = %(application_id)s',
        'pct_incidents_claimed_last_week': '''SELECT ROUND(COUNT(`owner_id`) / COUNT(*) * 100, 2)
                                               FROM `incident`
                                               WHERE `created` > %(date_variable)s - INTERVAL 1 WEEK
                                               AND `created` < %(date_variable)s
                                               AND `application_id` = %(application_id)s''',
        'median_seconds_to_claim_last_week': '''SELECT @incident_count := (SELECT count(*)
                                                                            FROM `incident`
                                                                            WHERE `created` > %(date_variable)s - INTERVAL 1 WEEK
                                                                            AND `created` < %(date_variable)s
                                                                            AND `active` = FALSE
                                                                            AND NOT ISNULL(`owner_id`)
                                                                            AND NOT ISNULL(`updated`)
                                                                            AND `application_id` = %(application_id)s),
                                                                    @row_id := 0,
                                                                    (SELECT CEIL(AVG(time_to_claim)) AS median
                                                                    FROM (SELECT `updated` - `created` AS time_to_claim
                                                                          FROM `incident`
                                                                          WHERE `created` > %(date_variable)s - INTERVAL 1 WEEK
                                                                          AND `created` < %(date_variable)s
                                                                          AND `active` = FALSE
                                                                          AND NOT ISNULL(`owner_id`)
                                                                          AND NOT ISNULL(`updated`)
                                                                          AND `application_id` = %(application_id)s
                                                                          ORDER BY time_to_claim) AS time_to_claim
                                                                    WHERE (SELECT @row_id := @row_id + 1)
                                                                    BETWEEN @incident_count/2.0 AND @incident_count/2.0 + 1)''',
        'total_call_retry_last_week': '''SELECT COUNT(*)
                                          FROM `twilio_retry` JOIN `message` ON `message`.`id` = `twilio_retry`.`message_id`
                                          WHERE `sent` > %(date_variable)s - INTERVAL 1 WEEK
                                          AND `sent` < %(date_variable)s
                                          AND `application_id` = %(application_id)s''',

        'high_priority_incidents_last_week': '''SELECT application.name, COUNT(DISTINCT incident.application_id, incident.id) as incident_count
                                                        FROM incident JOIN application ON application.id = %(application_id)s JOIN message ON message.incident_id = incident.id AND application.id = incident.application_id
                                                        WHERE incident.created > %(date_variable)s - INTERVAL 1 WEEK AND incident.created < %(date_variable)s AND priority_id IN (SELECT id FROM priority WHERE name IN ("high","urgent"))
                                                        GROUP BY incident.application_id ORDER BY incident_count DESC'''

    }

    mode_status = defaultdict(lambda: defaultdict(int))

    mode_stats_types = {
        'sms': {
            'fail': ['undelivered'],
            'success': ['sent', 'delivered']
        },
        'call': {
            'success': ['completed'],
            'fail': ['failed']
        },
        'email': {
            'success': [1],
            'fail': [0],
        }
    }

    cursor.execute('SELECT `name` FROM `mode`')
    modes = [row[0] for row in cursor]
    stats = {}
    fields = queries.keys()
    if fields_filter:
        fields &= set(fields_filter)

    for delta in range(0, 7):
        # 3600 seconds in an hour, round down to the nearest hour
        # 604800 seconds in a week, shift date by delta number of weeks
        unix_date = int(time.time() // 3600 * 3600) - (delta * 604800)
        now = datetime.fromtimestamp(unix_date)
        formatted_date = now.strftime('%Y-%m-%d %H:%M:%S')
        query_data = {'application_id': app['id'], 'date_variable': formatted_date}

        for key in fields:
            start = time.time()
            cursor.execute(queries[key], query_data)
            result = cursor.fetchone()
            if result:
                result = result[-1]

            logger.info('Stats query %s took %s seconds', key, round(time.time() - start, 2))

            # {statistic : {timestamp: value, timestamp: value}}
            if stats.get(key):
                stats[key].append({unix_date: result})
            else:
                stats[key] = []
                stats[key].append({unix_date: result})

        start = time.time()

        cursor.execute('''SELECT `mode`.`name`, COALESCE(`generic_message_sent_status`.`status`, `twilio_delivery_status`.`status`) AS thisStatus,
                            COUNT(*) FROM `message` USE INDEX FOR JOIN (ix_message_created)
                        LEFT JOIN `twilio_delivery_status` ON `twilio_delivery_status`.`message_id` = `message`.`id`
                        LEFT JOIN `generic_message_sent_status` ON `generic_message_sent_status`.`message_id` = `message`.`id`
                        JOIN `mode` ON `mode`.`id` = `message`.`mode_id`
                        WHERE ((NOT ISNULL(`twilio_delivery_status`.`status`) AND `mode`.`name` != 'email') OR (NOT ISNULL(`generic_message_sent_status`.`status`) AND `mode`.`name` = 'email'))
                            AND `message`.`application_id` = %(application_id)s
                            AND `message`.`created` > %(date_variable)s - INTERVAL 1 WEEK
                            AND `message`.`created` < %(date_variable)s
                        GROUP BY thisStatus, `mode`.`name`''', query_data)
        logger.info('App Stats (%s) mode status query took %s seconds', app['name'], round(time.time() - start, 2))

        for mode, status, count in cursor:
            if isinstance(status, str) and status.isdigit():
                status = int(status)
            mode_status[mode][status] = count

        for mode in mode_stats_types:
            status_counts = mode_status[mode]
            overall = float(sum(status_counts.values()))
            if overall > 0:
                fail_pct = round(
                    (sum(status_counts.pop(key, 0) for key in mode_stats_types[mode]['fail']) / overall) * 100,
                    2)
                success_pct = round(
                    (sum(status_counts.pop(key, 0) for key in mode_stats_types[mode]['success']) / overall) * 100, 2)
                other_pct = round(sum(status_counts.values()) / overall * 100, 2)
            else:
                fail_pct = success_pct = other_pct = None

            key = 'pct_%s_success_last_week' % mode
            if stats.get(key):
                stats['pct_%s_success_last_week' % mode].append({unix_date: success_pct})
                stats['pct_%s_fail_last_week' % mode].append({unix_date: fail_pct})
                stats['pct_%s_other_last_week' % mode].append({unix_date: other_pct})
            else:
                stats['pct_%s_success_last_week' % mode] = []
                stats['pct_%s_fail_last_week' % mode] = []
                stats['pct_%s_other_last_week' % mode] = []

                stats['pct_%s_success_last_week' % mode].append({unix_date: success_pct})
                stats['pct_%s_fail_last_week' % mode].append({unix_date: fail_pct})
                stats['pct_%s_other_last_week' % mode].append({unix_date: other_pct})

        # Get counts of messages sent per mode
        cursor.execute('''SELECT `mode`.`name`, `msg_count` FROM
                            (SELECT `mode_id`, COUNT(*) AS `msg_count` FROM `message`
                            USE INDEX (ix_message_sent)
                            WHERE `sent` > %(date_variable)s - INTERVAL 1 WEEK
                            AND `sent` < %(date_variable)s
                            AND `application_id` = %(application_id)s
                            GROUP BY `mode_id`) `counts`
                        JOIN `mode` ON `mode`.`id` = `counts`.`mode_id`''',
                       query_data)

        for row in cursor:
            key = 'total_%s_sent_last_week' % row[0]
            result = row[1]

            if stats.get(key):
                stats[key].append({unix_date: result})
            else:
                stats[key] = []
                stats[key].append({unix_date: result})

        # Zero out modes that don't show up in the count
        for mode in modes:
            count_stat = 'total_%s_sent_last_week' % mode
            if count_stat not in stats:
                stats[count_stat] = []
                stats[count_stat].append({unix_date: 0})
            elif unix_date not in stats[count_stat]:
                stats[count_stat].append({unix_date: 0})

    return stats


def calculate_single_stat(connection, cursor, stat_name):

    queries = {
        'total_incidents': 'SELECT application.name, COUNT(*) as cnt FROM `incident` JOIN application ON application.id = incident.application_id WHERE `created` < %(date_variable)s GROUP BY application.name ORDER BY cnt DESC, application.name ASC',
        'total_messages_sent': 'SELECT application.name, COUNT(*) as cnt FROM `message` JOIN application ON application.id = message.application_id WHERE `sent` < %(date_variable)s GROUP BY application.name ORDER BY cnt DESC, application.name ASC',
        'total_incidents_last_week': 'SELECT application.name, COUNT(*) as cnt FROM `incident` JOIN application ON application.id = incident.application_id WHERE `created` > %(date_variable)s - INTERVAL 1 WEEK AND `created` < %(date_variable)s GROUP BY application.name ORDER BY cnt DESC, application.name ASC',
        'total_messages_sent_last_week': 'SELECT application.name, COUNT(*) as cnt FROM `message` USE INDEX (ix_message_sent) JOIN application ON application.id = message.application_id WHERE `sent` > %(date_variable)s - INTERVAL 1 WEEK AND `sent` < %(date_variable)s GROUP BY application.name ORDER BY cnt DESC, application.name ASC',
        'pct_incidents_claimed_last_week': '''SELECT application.name, (ROUND(COUNT(`owner_id`) / COUNT(*) * 100, 2)) as cnt
                                               FROM `incident`
                                               JOIN application ON application.id = incident.application_id
                                               WHERE `created` > %(date_variable)s - INTERVAL 1 WEEK
                                               AND `created` < %(date_variable)s
                                               GROUP BY application.name ORDER BY cnt DESC, application.name ASC''',
        'total_call_retry_last_week': '''SELECT application.name, COUNT(*) as cnt
                                          FROM `twilio_retry` JOIN `message` ON `message`.`id` = `twilio_retry`.`message_id`
                                          JOIN application ON application.id = message.application_id
                                          WHERE `sent` > %(date_variable)s - INTERVAL 1 WEEK
                                          AND `sent` < %(date_variable)s GROUP BY application.name ORDER BY cnt DESC, application.name ASC''',
        'high_priority_incidents_last_week': '''SELECT application.name, COUNT(DISTINCT incident.application_id, incident.id) AS incident_count
                                                FROM incident JOIN application ON application.id = incident.application_id
                                                JOIN message ON message.incident_id = incident.id WHERE incident.created > %(date_variable)s - INTERVAL 1 WEEK
                                                AND incident.created < %(date_variable)s
                                                AND priority_id IN ( SELECT id FROM priority WHERE name IN ('high', 'urgent') )
                                                GROUP BY application.name ORDER BY incident_count DESC, application.name ASC'''

    }

    stats = {}

    for delta in range(0, 7):
        # 3600 seconds in an hour, round down to the nearest hour
        # 604800 seconds in a week, shift date by delta number of weeks
        unix_date = int(time.time() // 3600 * 3600) - (delta * 604800)
        now = datetime.fromtimestamp(unix_date)
        formatted_date = now.strftime('%Y-%m-%d %H:%M:%S')
        stats[unix_date] = []

        row_count = cursor.execute(queries[stat_name], {'date_variable': formatted_date})
        if row_count > 0:
            for app_name, count in cursor:
                stats[unix_date].append({app_name: count})

    return stats


def calculate_global_stats(connection, cursor, fields_filter=None):

    queries = {
        'total_plans': 'SELECT COUNT(*) FROM `plan` WHERE `created` <= %(date_variable)s',
        'total_incidents': 'SELECT COUNT(*) FROM `incident` WHERE `created` <= %(date_variable)s',
        'total_messages_sent': 'SELECT COUNT(*) FROM `message` WHERE `sent` <= %(date_variable)s',
        'total_incidents_last_week': 'SELECT COUNT(*) FROM `incident` WHERE `created` > %(date_variable)s - INTERVAL 1 WEEK AND `created` < %(date_variable)s',
        'total_messages_sent_last_week': 'SELECT COUNT(*) FROM `message` WHERE `sent` > %(date_variable)s - INTERVAL 1 WEEK AND `sent` < %(date_variable)s',
        'total_active_users': 'SELECT COUNT(*) FROM `target` WHERE `type_id` = (SELECT `id` FROM `target_type` WHERE `name` = "user") AND `active` = TRUE',
        'pct_incidents_claimed_last_week': '''SELECT ROUND(
                                                (SELECT COUNT(*) FROM `incident`
                                                WHERE `created` > (%(date_variable)s - INTERVAL 1 WEEK)
                                                AND `created` < (%(date_variable)s)
                                                AND `active` = FALSE
                                                AND NOT isnull(`owner_id`)) /
                                                (SELECT COUNT(*) FROM `incident`
                                                WHERE `created` > (%(date_variable)s - INTERVAL 1 WEEK)
                                                AND `created` < (%(date_variable)s)) * 100, 2)''',
        'median_seconds_to_claim_last_week': '''SELECT @incident_count := (SELECT count(*)
                                                                            FROM `incident`
                                                                            WHERE `created` > (%(date_variable)s - INTERVAL 1 WEEK)
                                                                            AND `created` < (%(date_variable)s)
                                                                            AND `active` = FALSE
                                                                            AND NOT ISNULL(`owner_id`)
                                                                            AND NOT ISNULL(`updated`)),
                                                        @row_id := 0,
                                                        (SELECT CEIL(AVG(time_to_claim)) as median
                                                        FROM (SELECT `updated` - `created` as time_to_claim
                                                                FROM `incident`
                                                                WHERE `created` > (%(date_variable)s - INTERVAL 1 WEEK)
                                                                AND `created` < (%(date_variable)s)
                                                                AND `active` = FALSE
                                                                AND NOT ISNULL(`owner_id`)
                                                                AND NOT ISNULL(`updated`)
                                                                ORDER BY time_to_claim) as time_to_claim
                                                        WHERE (SELECT @row_id := @row_id + 1)
                                                        BETWEEN @incident_count/2.0 AND @incident_count/2.0 + 1)''',
        'total_applications': 'SELECT COUNT(*) FROM `application` WHERE `auth_only` = FALSE',
        'total_high_priority_incidents_last_week': 'SELECT COUNT(DISTINCT incident.id) FROM incident JOIN message ON message.incident_id = incident.id WHERE incident.created > %(date_variable)s - INTERVAL 1 WEEK AND incident.created < %(date_variable)s AND priority_id IN (SELECT id FROM priority WHERE name IN ("high","urgent"))'
    }

    stats = {}
    fields = queries.keys()
    if fields_filter:
        fields &= set(fields_filter)

    for delta in range(0, 7):

        # 3600 seconds in an hour, round down to the nearest hour
        # 604800 seconds in a week, shift date by delta number of weeks
        unix_date = int(time.time() // 3600 * 3600) - (delta * 604800)
        now = datetime.fromtimestamp(unix_date)
        formatted_date = now.strftime('%Y-%m-%d %H:%M:%S')

        for key in fields:
            start = time.time()
            cursor.execute(queries[key], {'date_variable': formatted_date})
            result = cursor.fetchone()
            if result:
                result = result[-1]

            logger.info('Stats query %s took %s seconds', key, round(time.time() - start, 2))

            # format: {statistic : [{timestamp: value}, {timestamp: value}]}
            if key in stats:
                stats[key].append({unix_date: result})
            else:
                stats[key] = []
                stats[key].append({unix_date: result})

    return stats



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import pytest
from falcon import HTTPBadRequest


def test_parse_valid_body():
    from iris.webhooks.grafana import grafana
    grafana_webhook = grafana()

    fake_post = {
        "evalMatches": [{
            "value": 100,
            "metric": "High value",
            "tags": "",
        }, {
            "value": 200,
            "metric": "Higher Value",
            "tags": "",
        }],
        "imageUrl": "http://grafana.org/assets/img/blog/mixed_styles.png",
        "message": "Someone is testing the alert notification within grafana.",
        "ruleId": 0,
        "ruleName": "Test notification",
        "ruleUrl": "https://grafana.org/",
        "state": "alerting",
        "title": "[Alerting] Test notification"
    }
    grafana_webhook.validate_post(fake_post)


def test_parse_invalid_body():
    from iris.webhooks.grafana import grafana
    grafana_webhook = grafana()

    fake_post = {
        "evalMatches": [{
            "value": 100,
            "metric": "High value",
            "tags": "",
        }, {
            "value": 200,
            "metric": "Higher Value",
            "tags": "",
        }],
        "imageUrl": "http://grafana.org/assets/img/blog/mixed_styles.png",
        "message": "Someone is testing the alert notification within grafana.",
        "ruleId": 0,
        "ruleName": "Test notification",
        "ruleUrl": "https://grafana.org/",
        "title": "[Alerting] Test notification"
    }

    with pytest.raises(HTTPBadRequest):
        grafana_webhook.validate_post(fake_post)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from iris.custom_import import import_custom_module

_registered_validators = []


def init_validators(validators):
    for module in validators:
        _registered_validators.append(import_custom_module('iris.validators', module)())


class IrisValidationException(Exception):
    pass


def run_validation(item, *args, **kwargs):
    for validator in _registered_validators:
        try:
            getattr(validator, 'validate_' + item)(*args, **kwargs)
        except RuntimeError as e:
            raise IrisValidationException(e)



from iris.bin.owasync import is_pointless_message, relay, default_metrics
from exchangelib.items import Message
from exchangelib.properties import MessageHeader, Mailbox
from iris import metrics
from iris.client import IrisClient
import requests_mock


def test_is_pointless_message():
    bad_message_headers = [{
        'name': 'X-Autoreply',
        'value': 'yes'
    }]

    bad_message2_headers = [
        {
            'name': 'Auto-Submitted',
            'value': 'auto-replied'
        },
        {
            'name': 'Precedence',
            'value': 'bulk'
        }
    ]

    good_message_headers = [{
        'name': 'From',
        'value': 'Foo Bar <foo@bar.com>'
    }]

    assert is_pointless_message(bad_message_headers)
    assert is_pointless_message(bad_message2_headers)
    assert not is_pointless_message(good_message_headers)


def test_metrics():
    metrics.stats.update(default_metrics)
    bad_message = Message(headers=None, sender=Mailbox(email_address=''))
    relay(bad_message, None)
    assert metrics.stats['message_ignore_count'] == 1

    metrics.stats.update(default_metrics)
    bad_message = Message(headers=[MessageHeader(name='X-Autoreply', value='yes')], sender=Mailbox(email_address=''))
    relay(bad_message, None)
    assert metrics.stats['message_ignore_count'] == 1

    fake_iris_client = IrisClient('http://localhost')
    good_message = Message(headers=[], sender=Mailbox(email_address=''), text_body='')

    with requests_mock.mock() as m:
        m.post('http://localhost/v0/response/email')
        metrics.stats.update(default_metrics)
        relay(good_message, fake_iris_client)
        assert metrics.stats['message_relay_success_count'] == 1
        assert metrics.stats['incident_created_count'] == 0

    with requests_mock.mock() as m:
        m.post('http://localhost/v0/response/email', headers={'X-IRIS-INCIDENT': '1234'})
        metrics.stats.update(default_metrics)
        relay(good_message, fake_iris_client)
        assert metrics.stats['message_relay_success_count'] == 1
        assert metrics.stats['incident_created_count'] == 1

    with requests_mock.mock() as m:
        m.post('http://localhost/v0/response/email', status_code=500)
        metrics.stats.update(default_metrics)
        relay(good_message, fake_iris_client)
        assert metrics.stats['message_relay_failure_count'] == 1

    with requests_mock.mock() as m:
        m.post('http://localhost/v0/response/email', status_code=400)
        metrics.stats.update(default_metrics)
        relay(good_message, fake_iris_client)
        assert metrics.stats['malformed_message_count'] == 1



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from iris import db
import logging
logger = logging.getLogger(__name__)


class mailing_list(object):
    def __init__(self, config):
        self.max_list_names = config.get('ldap_lists', {}).get('max_unrolled_users', 0)

    def get(self, role, target):
        if role == 'mailing-list':
            return self.unroll_mailing_list(target)
        else:
            return None

    def unroll_mailing_list(self, list_name):
        connection = db.engine.raw_connection()
        cursor = connection.cursor()

        cursor.execute('''
          SELECT `mailing_list`.`target_id`,
                 `mailing_list`.`count`
          FROM `mailing_list`
          JOIN `target` on `target`.`id` = `mailing_list`.`target_id`
          WHERE `target`.`name` = %s
        ''', list_name)

        list_info = cursor.fetchone()

        if not list_info:
            logger.warning('Invalid mailing list %s', list_name)
            cursor.close()
            connection.close()
            return None

        list_id, list_count = list_info

        if self.max_list_names > 0 and list_count >= self.max_list_names:
            logger.warning('Not returning any results for list group %s as it contains too many members (%s > %s)',
                           list_name, list_count, self.max_list_names)
            cursor.close()
            connection.close()
            return None

        cursor.execute('''SELECT `target`.`name`
                          FROM `mailing_list_membership`
                          JOIN `target` on `target`.`id` = `mailing_list_membership`.`user_id`
                          WHERE `mailing_list_membership`.`list_id` = %s''', [list_id])
        names = [row[0] for row in cursor]

        cursor.close()
        connection.close()

        logger.info('Unfurled %s people from list %s', len(names), list_name)
        return names



from importlib import import_module


def import_custom_module(default_root, module):
    if '.' in module:
        module_path = module
        module = module.split('.')[-1]
    else:
        module_path = default_root + '.' + module
    return getattr(import_module(module_path), module)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from itertools import cycle
import logging
logger = logging.getLogger(__name__)


class Coordinator():
    def __init__(self, is_master, slaves=[]):
        self.is_master = is_master
        if is_master:
            logger.info('I am the master sender')
        else:
            logger.info('I am a slave sender')

        self.slaves = cycle((slave['host'], slave['port']) for slave in slaves)
        self.slave_count = len(slaves)

    def update_forever(self):
        pass

    def leave_cluster(self):
        pass

    def am_i_master(self):
        return self.is_master

    def get_current_master(self):
        return None



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from gevent import monkey, sleep, spawn, queue
monkey.patch_all()  # NOQA

import logging
import logging.handlers
import time
import ujson
import os
import socket
import gevent
import signal
import setproctitle
import copy

from collections import defaultdict
from iris.plugins import init_plugins
from iris.vendors import IrisVendorManager, iris_smtp
from iris.sender import auditlog
from iris import metrics
from uuid import uuid4
from iris.gmail import Gmail
from iris import db
from iris.api import load_config
from iris.utils import sanitize_unicode_dict
from iris.sender import rpc, cache
from iris.sender.message import update_message_mode
from iris.sender.oneclick import oneclick_email_markup, generate_oneclick_url
from iris import cache as api_cache
from iris.sender.quota import ApplicationQuota
from iris.role_lookup import IrisRoleLookupException
from pymysql import DataError
# queue for sending messages
from iris.sender.shared import per_mode_send_queues, add_mode_stat

# sql

NEW_INCIDENTS = '''SELECT
    `incident`.`id` as `id`,
    `incident`.`created` as `created`,
    `incident`.`plan_id` as `plan_id`,
    `incident`.`context` as `context`,
    `application`.`name` as `application`
FROM `incident`
JOIN `application`
ON `incident`.`application_id`=`application`.`id`
WHERE `current_step`=0 AND `active`=1'''

INACTIVE_SQL = '''UPDATE
`incident`
SET `active`=0
WHERE `id` IN %s'''

GET_INACTIVE_IDS_SQL = '''SELECT
distinct `incident_id`
FROM (
    SELECT
        `incident_id`,
        `max`,
        `age`,
        MAX(`count`) as max_count
        FROM (
            SELECT
            `message`.`incident_id` as `incident_id`,
            `message`.`plan_notification_id` as `plan_notification_id`,
            count(`message`.`id`) as `count`,
            `plan_notification`.`repeat` + 1 as `max`,
            TIMESTAMPDIFF(SECOND, MAX(`message`.`sent`), NOW()) as `age`,
            `plan_notification`.`wait` as `wait`,
            `plan_notification`.`step` as `step`,
            `incident`.`current_step`,
            `plan`.`step_count`,
            `message`.`plan_id`,
            `message`.`application_id`,
            `incident`.`context`
            FROM `message`
            JOIN `incident` ON `message`.`incident_id` = `incident`.`id`
            JOIN `plan_notification` ON `message`.`plan_notification_id` = `plan_notification`.`id`
            JOIN `plan` ON `message`.`plan_id` = `plan`.`id`
            WHERE `incident`.`active` = 1
            AND `incident`.`current_step`=`plan`.`step_count`
            AND `step` = `incident`.`current_step`
            GROUP BY `incident`.`id`, `message`.`plan_notification_id`, `message`.`target_id`
        ) as `inner`
        GROUP BY `incident_id`, `plan_notification_id`
        HAVING `max_count` = `max` AND BIT_AND(`age` > `wait`) = 1
    ) as `exhausted_incidents`'''

QUEUE_SQL = '''SELECT
`incident_id`,
`plan_id`,
`plan_notification_id`,
max(`count`) as `count`,
`max`,
`age`,
`wait`,
`step`,
`current_step`,
`step_count`
FROM (
    SELECT
        `message`.`incident_id` as `incident_id`,
        `message`.`plan_notification_id` as `plan_notification_id`,
        count(`message`.`id`) as `count`,
        `plan_notification`.`repeat` + 1 as `max`,
        TIMESTAMPDIFF(SECOND, max(`message`.`created`), NOW()) as `age`,
        `plan_notification`.`wait` as `wait`,
        `plan_notification`.`step` as `step`,
        `incident`.`current_step`,
        `plan`.`step_count`,
        `message`.`plan_id`,
        `message`.`application_id`,
        `incident`.`context`
    FROM `message`
    JOIN `incident` ON `message`.`incident_id` = `incident`.`id`
    JOIN `plan_notification` ON `message`.`plan_notification_id` = `plan_notification`.`id`
    JOIN `plan` ON `message`.`plan_id` = `plan`.`id`
    WHERE `incident`.`active` = 1
    GROUP BY `incident`.`id`, `message`.`plan_notification_id`, `message`.`target_id`
) as `inner`
GROUP BY `incident_id`, `plan_notification_id`
HAVING `age` > `wait` AND (`count` < `max`
                           OR (`count` = `max` AND `step` = `current_step`
                               AND `step` < `step_count`))'''

UPDATE_INCIDENT_SQL = '''UPDATE `incident` SET `current_step`=%s WHERE `id`=%s'''

INVALIDATE_INCIDENT = '''UPDATE `incident` SET `active`=0 WHERE `id`=%s'''

INSERT_MESSAGE_SQL = '''INSERT INTO `message`
    (`created`, `plan_id`, `plan_notification_id`, `incident_id`, `application_id`, `target_id`, `priority_id`, `body`)
VALUES (NOW(), %s,%s,%s,%s,%s,%s,%s)'''

UNSENT_MESSAGES_SQL = '''SELECT
    `msg`.`body`,
    `msg`.`id` as `message_id`,
    `target`.`name` as `target`,
    `priority`.`name` as `priority`,
    `priority`.`id` as `priority_id`,
    `application`.`name` as `application`,
    `plan`.`name` as `plan`,
    `plan`.`id` as `plan_id`,
    `incident`.`id` as `incident_id`,
    `incident`.`context` as `context`,
    `incident`.`created` as `incident_created`,
    `plan_notification`.`template` as `template`,
    `dynamic_target`.`name` as `dynamic_target`
FROM (SELECT * FROM `message` FORCE INDEX (ix_message_active) WHERE `active`=1) AS msg
JOIN `application` ON `msg`.`application_id`=`application`.`id`
JOIN `priority` ON `msg`.`priority_id`=`priority`.`id`
LEFT OUTER JOIN `target` ON `msg`.`target_id`=`target`.`id`
LEFT OUTER JOIN `plan` ON `msg`.`plan_id`=`plan`.`id`
LEFT OUTER JOIN `plan_notification` ON `msg`.`plan_notification_id`=`plan_notification`.`id`
LEFT OUTER JOIN `incident` ON `msg`.`incident_id`=`incident`.`id`
LEFT OUTER JOIN `dynamic_plan_map` ON `incident`.`id` = `dynamic_plan_map`.`incident_id`
  AND `plan_notification`.`dynamic_index` = `dynamic_plan_map`.`dynamic_index`
LEFT OUTER JOIN `target` `dynamic_target` ON `dynamic_target`.`id` = `dynamic_plan_map`.`target_id`'''

SENT_MESSAGE_BATCH_SQL = '''UPDATE `message`
SET `destination`=%%s,
    `mode_id`=%%s,
    `template_id`=%%s,
    `batch`=%%s,
    `active`=FALSE,
    `sent`=NOW()
WHERE `id` IN %s'''

SENT_MESSAGE_SQL = '''UPDATE `message`
SET `destination`=%s,
    `mode_id`=%s,
    `template_id`=%s,
    `active`=FALSE,
    `sent`=NOW()
WHERE `id`=%s'''

UPDATE_MESSAGE_BODY_SQL = '''UPDATE `message`
                             SET `body`=%s,
                                 `subject`=%s
                             WHERE `id` IN %s'''

PRUNE_OLD_AUDIT_LOGS_SQL = '''DELETE FROM `message_changelog` WHERE `date` < DATE_SUB(CURDATE(), INTERVAL 3 MONTH)'''

# When a rendered message body is longer than this number of characters, drop it.
MAX_MESSAGE_BODY_LENGTH = 40000
MAX_MESSAGE_RETRIES = 2

# logging
logger = logging.getLogger()
formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s %(message)s')
log_file = os.environ.get('SENDER_LOG_FILE')
if log_file:
    ch = logging.handlers.RotatingFileHandler(log_file, mode='a', maxBytes=10485760, backupCount=10)
else:
    ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logger.setLevel(logging.INFO)
logger.addHandler(ch)

pidfile = os.environ.get('SENDER_PIDFILE')
if pidfile:
    try:
        pid = os.getpid()
        with open(pidfile, 'w') as h:
            h.write('%s\n' % pid)
            logger.info('Wrote pid %s to %s', pid, pidfile)
    except IOError:
        logger.exception('Failed writing pid to %s', pidfile)


# rate limiting data structure message key -> minute -> count
# used to calcuate if a new message exceeds the rate limit
# and needs to be queued
plan_aggregate_windows = {}

# all messages waiting to be queue across all keys
messages = {}

# queue of message_ids waiting to be sent for a given key
queues = {}

# time last message was queued while in aggregation mode for a given key
# used to determine if a new message should be aggregated or if aggregation should end
aggregation = {}

# last time a batch was sent for a given key
# used to determine if it's time to send the next batch
sent = {}

# queue for messages entering the system
# this sets the ground work for not having to poll the DB for messages
message_queue = queue.Queue()

# Quota object used for rate limiting
quota = None

# Coordinator object for sender master election
coordinator = None

# Mode -> [{'greenlet': greenlet, 'kill_set': gevent.Event}]
worker_tasks = defaultdict(list)

# MX -> [{'greenlet': greenlet, 'kill_set': gevent.Event}]
autoscale_email_worker_tasks = defaultdict(list)

# support the 2nd control+c force exiting sender without waiting for tasks to finish
shutdown_started = False

# Set of active message IDs currently being sent, to avoid re-sending messages that are currently
# just blocked on a downstream such as smtp
message_ids_being_sent = set()


default_sender_metrics = {
    'email_cnt': 0, 'email_total': 0, 'email_fail': 0, 'email_sent': 0, 'email_max': 0,
    'email_min': 0, 'email_avg': 0, 'im_cnt': 0, 'im_total': 0, 'im_fail': 0, 'im_sent': 0,
    'im_max': 0, 'im_min': 0, 'im_avg': 0, 'slack_cnt': 0, 'slack_total': 0, 'slack_fail': 0,
    'slack_sent': 0, 'slack_max': 0, 'slack_min': 0, 'slack_avg': 0, 'drop_cnt': 0,
    'drop_total': 0, 'drop_fail': 0, 'drop_sent': 0, 'drop_max': 0, 'drop_min': 0, 'drop_avg': 0,
    'sms_cnt': 0, 'sms_total': 0, 'sms_fail': 0, 'sms_sent': 0, 'sms_max': 0,
    'sms_min': 0, 'sms_avg': 0, 'call_cnt': 0, 'call_total': 0,
    'call_fail': 0, 'call_sent': 0, 'call_max': 0, 'call_min': 0, 'call_avg': 0, 'task_failure': 0,
    'oncall_error': 0, 'role_target_lookup_error': 0, 'target_not_found': 0, 'message_send_cnt': 0,
    'notification_cnt': 0, 'api_request_cnt': 0, 'api_request_timeout_cnt': 0,
    'rpc_message_pass_success_cnt': 0, 'rpc_message_pass_fail_cnt': 0,
    'slave_message_send_success_cnt': 0, 'slave_message_send_fail_cnt': 0,
    'msg_drop_length_cnt': 0, 'send_queue_gets_cnt': 0, 'send_queue_puts_cnt': 0, 'send_queue_puts_fail_cnt': 0,
    'send_queue_email_size': 0, 'send_queue_im_size': 0, 'send_queue_slack_size': 0, 'send_queue_call_size': 0,
    'send_queue_sms_size': 0, 'send_queue_drop_size': 0, 'new_incidents_cnt': 0, 'workers_respawn_cnt': 0,
    'message_retry_cnt': 0, 'message_ids_being_sent_cnt': 0, 'notifications': 0, 'deactivation': 0,
    'new_msg_count': 0, 'poll': 0, 'queue': 0, 'aggregations': 0, 'hipchat_cnt': 0, 'hipchat_fail': 0,
    'hipchat_total': 0, 'hipchat_sent': 0, 'hipchat_max': 0, 'hipchat_min': 0
}

# TODO: make this configurable
target_fallback_mode = 'email'
should_mock_gwatch_renewer = False
config = None


# msg_info takes the form [(incident_id, plan_notification_id), ...]
def create_messages(msg_info):
    msg_count = 0
    error_incident_ids = set()
    connection = db.engine.raw_connection()
    cursor = connection.cursor()

    for (incident_id, plan_notification_id) in msg_info:
        application_id = cache.incidents[incident_id]['application_id']
        plan_notification = cache.plan_notifications[plan_notification_id]
        if plan_notification['role_id'] is None and plan_notification['target_id'] is None:
            dynamic_info = cache.dynamic_plan_map[incident_id][plan_notification['dynamic_index']]
            role = cache.roles[dynamic_info['role_id']]['name']
            target = cache.targets[dynamic_info['target_id']]['name']
        else:
            role = cache.roles[plan_notification['role_id']]['name']
            target = cache.targets[plan_notification['target_id']]['name']

        # find role/priority from plan_notification_id
        try:
            names = cache.targets_for_role(role, target)
        except IrisRoleLookupException as e:
            names = None
            metrics.incr('role_target_lookup_error')
            lookup_fail_reason = str(e)
        else:
            lookup_fail_reason = None

        priority_id = plan_notification['priority_id']
        redirect_to_plan_owner = False
        body = ''

        if not names:

            # if message is optional don't bother the creator, simply return true instead
            if plan_notification['optional']:
                msg_count += 1
                continue

            # Try to get creator of the plan and nag them instead
            name = None
            try:
                name = cache.plans[plan_notification['plan_id']]['creator']
            except (KeyError, TypeError):
                pass

            if not name:
                logger.error(('Failed to find targets for incident %s, plan_notification_id: %s, '
                              'role: %s, target: %s, result: %s and failed looking '
                              'up the plan\'s creator'),
                             incident_id, plan_notification_id, role, target, names)
                error_incident_ids.add(incident_id)
                continue

            try:
                priority_id = api_cache.priorities['low']['id']
            except KeyError:
                logger.error(('Failed to find targets for incident %s, plan_notification_id: %s, '
                              'role: %s, target: %s, result: %s and failed looking '
                              'up ID for low priority'),
                             incident_id, plan_notification_id, role, target, names)
                error_incident_ids.add(incident_id)
                continue

            logger.error(('Failed to find targets for incident %s, plan_notification_id: %s, '
                          'role: %s, target: %s, result: %s. '
                          'Reaching out to %s instead and lowering priority to low (%s)'),
                         incident_id, plan_notification_id, role, target, names, name, priority_id)

            body = ('You are receiving this as you created this plan and we can\'t resolve'
                    ' %s of %s at this time%s.\n\n') % (role, target, ': %s' % lookup_fail_reason if lookup_fail_reason else '')

            names = [name]
            redirect_to_plan_owner = True

        for name in names:
            t = cache.target_names[name]
            if t:
                target_id = t['id']
                retries = 0
                max_retries = 5
                while True:
                    retries += 1
                    try:
                        cursor.execute(INSERT_MESSAGE_SQL,
                                       (plan_notification['plan_id'], plan_notification_id, incident_id,
                                        application_id, target_id, priority_id, body))
                        connection.commit()
                    except Exception:
                        logger.warning('Failed inserting message for incident %s. (Try %s/%s)', incident_id, retries, max_retries)
                        if retries < max_retries:
                            sleep(.2)
                            continue
                        else:
                            raise Exception('Failed inserting message retries exceeded')
                    else:
                        if redirect_to_plan_owner:
                            # needed for the lastrowid to exist in the DB to satisfy the constraint
                            auditlog.message_change(
                                cursor.lastrowid,
                                auditlog.TARGET_CHANGE,
                                role + '|' + target,
                                name,
                                lookup_fail_reason or 'Changing target to plan owner as we failed resolving original target')
                        break

                msg_count += 1
            else:
                metrics.incr('target_not_found')
                logger.warning('Failed to notify plan creator; no active target found: %s', name)
    cursor.close()
    connection.close()
    return msg_count, error_incident_ids


def deactivate():
    # deactivate incidents that have expired
    logger.info('[-] start deactivate task...')
    start_deactivation = time.time()

    connection = db.engine.raw_connection()
    cursor = connection.cursor()

    max_retries = 3

    # this deadlocks sometimes. try until it doesn't.
    for i in range(1, max_retries + 1):
        try:
            cursor.execute(GET_INACTIVE_IDS_SQL)
            ids = tuple(r[0] for r in cursor)
            if ids:
                cursor.execute(INACTIVE_SQL, (ids,))
                connection.commit()
                break
        except Exception:
            if i == max_retries:
                logger.warning('Failed running deactivate query. (Try %s/%s)', i, max_retries)
            else:
                logger.warning('Deadlocked running deactivate query. (Try %s/%s)', i, max_retries)
            sleep(.2)

    cursor.close()
    connection.close()

    metrics.set('deactivation', time.time() - start_deactivation)
    logger.info('[*] deactivate task finished')


def escalate():
    # make notifications for things that should repeat or escalate
    logger.info('[-] start escalate task...')

    # first, handle new incidents
    start_notifications = time.time()

    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute(NEW_INCIDENTS)

    escalations = {}
    incident_per_plan_cnt = {}
    for incident_id, created, plan_id, context, application in cursor:
        incident_per_plan_cnt[plan_id] = incident_per_plan_cnt.get(plan_id, 0) + 1
        escalations[incident_id] = (plan_id, 1)
        # create tracking message if configured
        plan = cache.plans[plan_id]
        tracking_type = plan['tracking_type']
        tracking_key = plan['tracking_key']
        tracking_template = plan['tracking_template']
        app_tracking_template = tracking_template.get(application) if tracking_template else None
        if tracking_type and tracking_key and app_tracking_template:
            # plan defines tracking notifications
            context = ujson.loads(context)
            context['iris'] = {
                'incident_id': incident_id,
                'plan': plan['name'],
                'plan_id': plan_id,
                'application': application,
                'incident_created': created,
            }
            if tracking_type == 'call':
                # in case there are any plans with call tracking type that have not been cleaned up
                logger.warning('Tracking messages do not support support call mode, ignoring tracking message')
                continue
            if tracking_type == 'email':
                tracking_message = {
                    'noreply': True,
                    'destination': tracking_key,
                    'mode': tracking_type
                }

                try:
                    subject = app_tracking_template['email_subject'].render(**context)
                except Exception as e:
                    subject = 'plan %s - tracking notification subject failed to render: %s' % (plan['name'], e)
                    logger.exception(subject)
                tracking_message['email_subject'] = subject

                try:
                    body = app_tracking_template['email_text'].render(**context)
                except Exception as e:
                    body = 'plan %s - tracking notification body failed to render: %s' % (plan['name'], e)
                    logger.exception(body)
                tracking_message['email_text'] = body

                email_html_tpl = app_tracking_template.get('email_html')
                if email_html_tpl:
                    try:
                        html_body = email_html_tpl.render(**context)
                    except Exception as e:
                        html_body = 'plan %s - tracking notification html body failed to render: %s' % (plan['name'], e)
                        logger.exception(html_body)
                    tracking_message['email_html'] = html_body
            else:
                tracking_message = {
                    'noreply': True,
                    'destination': tracking_key,
                    'mode': tracking_type
                }
                try:
                    body = app_tracking_template['body'].render(**context)
                except Exception as e:
                    body = 'plan %s - tracking notification body failed to render: %s' % (plan['name'], e)
                    logger.exception(body)
                tracking_message['body'] = body

            message_send_enqueue(tracking_message)
    cursor.close()

    new_incidents_count = len(escalations)
    metrics.set('new_incidents_cnt', new_incidents_count)
    logger.info('[*] %s new incidents', new_incidents_count)

    # log any plan that creates a high volume of incidents
    for pln_id, cnt in incident_per_plan_cnt.items():
        if cnt > 10:
            logger.info("plan with id %d created %d incidents this loop iteration")

    # then, fetch message count for current incidents
    msg_count = 0
    cursor = connection.cursor(db.dict_cursor)
    cursor.execute(QUEUE_SQL)
    msg_info = []
    for n in cursor.fetchall():
        if n['count'] < n['max']:
            msg_info.append((n['incident_id'], n['plan_notification_id']))
        else:
            escalations[n['incident_id']] = (n['plan_id'], n['current_step'] + 1)
    msg_count += create_messages(msg_info)[0]

    # Create escalation messages
    msg_info = []
    for incident_id, (plan_id, step) in escalations.items():
        plan = cache.plans[plan_id]
        steps = plan['steps'].get(step, [])
        for plan_notification_id in steps:
            msg_info.append((incident_id, plan_notification_id))
    count, error_incident_ids = create_messages(msg_info)
    msg_count += count

    # Update incident step value
    for incident_id, (plan_id, step) in escalations.items():
        plan = cache.plans[plan_id]
        steps = plan['steps'].get(step, [])
        retries = 0
        max_retries = 5
        while True:
            retries += 1
            try:
                if steps:
                    if step == 1 and incident_id in error_incident_ids:
                        # no message created due to role look up failure, reset step to
                        # 0 for retry
                        step = 0
                    cursor.execute(UPDATE_INCIDENT_SQL, (step, incident_id))
                else:
                    logger.error('plan id %d has no steps, incident id %d is invalid', plan_id, incident_id)
                    cursor.execute(INVALIDATE_INCIDENT, incident_id)

                connection.commit()
            except Exception:
                logger.warning('Failed updating incident %s. (Try %s/%s)', incident_id, retries, max_retries)
                if retries < max_retries:
                    sleep(.2)
                    continue
                else:
                    raise Exception('Failed updating batch messages retries exceeded')
            else:
                break
    cursor.close()
    connection.close()

    logger.info('[*] %s new messages', msg_count)
    logger.info('[*] escalate task finished')
    metrics.set('notifications', time.time() - start_notifications)


def aggregate(now):
    # see if it's time to send the batches
    logger.info('[-] start aggregate task - queued: %s', len(messages))
    start_aggregations = time.time()
    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    cursor.execute('SELECT `id` FROM `message` WHERE active=1')
    all_actives = {r[0] for r in cursor}
    cursor.close()
    connection.close()
    for key in list(queues.keys()):
        aggregation_window = cache.plans[key[0]].get('aggregation_window')
        if aggregation_window is None:
            logger.error('No aggregation window found for plan %s', key[0])
            aggregation_window = 0
        if now - sent.get(key, 0) >= aggregation_window:
            aggregated_message_ids = queues[key]
            active_message_ids = aggregated_message_ids & all_actives
            l = len(active_message_ids)

            if l == 1:
                m = messages.pop(next(iter(active_message_ids)))
                logger.info('aggregate - %(message_id)s pushing to send queue', m)
                message_send_enqueue(m)
            elif l > 1:
                uuid = uuid4().hex
                m = messages[next(iter(active_message_ids))]
                logger.info('aggregate - %s pushing to send queue', uuid)
                m['batch_id'] = uuid

                # Cast from set to list, as sets are not msgpack serializable
                m['aggregated_ids'] = list(active_message_ids)
                message_send_enqueue(m)

                for message_id in aggregated_message_ids:
                    messages.pop(message_id, None)
                logger.info('[-] purged %s from messages %s remaining', active_message_ids, len(messages))
            del queues[key]
            sent[key] = now
    inactive_message_ids = messages.keys() - all_actives
    logger.info('[x] dropped %s inactive messages from claimed incidents, %s remain',
                len(inactive_message_ids), len(messages))

    # remove inactive message from the queue
    for message_id in inactive_message_ids:
        messages.pop(message_id, None)

    metrics.set('aggregations', time.time() - start_aggregations)
    logger.info('[*] aggregate task finished - queued: %s', len(messages))


def poll():
    # poll unsent messages
    logger.info('[-] start send task...')
    start_send = time.time()

    connection = db.engine.raw_connection()
    cursor = connection.cursor(db.dict_cursor)
    if messages:
        cursor.execute(UNSENT_MESSAGES_SQL + ' AND `msg`.`id` NOT IN %s', [tuple(messages)])
    else:
        cursor.execute(UNSENT_MESSAGES_SQL)

    new_msg_count = cursor.rowcount
    queued_msg_cnt = len(messages)
    metrics.set('new_msg_count', new_msg_count)
    logger.info('%d new messages waiting in database - queued: %d', new_msg_count, queued_msg_cnt)

    for m in cursor:
        # Set dynamic target as target if applicable
        if m.get('dynamic_target') and m.get('target') is None:
            m['target'] = m['dynamic_target']
        # iris's own email response does not have context since content and
        # subject are already set
        if m.get('context'):
            context = ujson.loads(m['context'])
            # inject meta variables
            context['iris'] = {k: m[k] for k in m if k != 'context'}
            m['context'] = context
        message_queue.put(m)

    metrics.set('poll', time.time() - start_send)
    metrics.set('queue', len(messages))
    logger.info('[*] send task finished')
    cursor.close()
    connection.close()


def fetch_and_prepare_message():
    now = time.time()
    m = message_queue.get()
    message_id = m['message_id']
    plan_id = m['plan_id']
    if plan_id is None:
        message_send_enqueue(m)
        return

    plan = cache.plans[plan_id]

    # queue key
    key = (m['plan_id'], m['application'], m['priority'], m['target'])

    # should this message be aggregated?
    aggregate = False
    last_aggregation = aggregation.get(key)
    if last_aggregation:
        if now - last_aggregation > plan['aggregation_reset']:
            # it's been long enough since the last message
            # return to immediate sending mode and clear aggregations
            del aggregation[key]
            try:
                # if we have sent batches before delete the entry for last sent
                del sent[key]
            except KeyError:
                pass
        else:
            # still getting enough messages fast enough to remain in aggregation
            aggregation[key] = now
            aggregate = True

    if aggregate:
        # we are still in a previous aggregation mode
        queues.setdefault(key, set()).add(message_id)
        messages[message_id] = m
    else:
        # does this message trigger aggregation?
        window = plan_aggregate_windows.setdefault(key, defaultdict(int))

        for bucket in list(window.keys()):
            if now - bucket > plan['threshold_window']:
                del window[bucket]

        window[now] += 1

        if sum(window.values()) > plan['threshold_count']:
            # too many messages for the aggregation key - enqueue

            # add message id to aggregation queue
            queues.setdefault(key, set()).add(message_id)
            # add message id to queue for deduping
            messages[message_id] = m
            # initialize last sent tracker
            sent[key] = now
            # initialize aggregation indicator
            aggregation[key] = now
            # TODO: also render message content here?
            audit_msg = 'Aggregated with key (%s, %s, %s, %s)' % key
            spawn(auditlog.message_change, m['message_id'], auditlog.SENT_CHANGE, '', '', audit_msg)
        else:
            # cleared for immediate sending
            message_send_enqueue(m)


def send():
    logger.info('[-] start send loop...')
    while True:
        fetch_and_prepare_message()
    logger.info('[*] send loop finished...')


def set_target_fallback_mode(message):
    try:
        connection = db.engine.raw_connection()
        cursor = connection.cursor()
        cursor.execute('''SELECT `destination`, `mode`.`name`, `mode`.`id`
                          FROM `target`
                          JOIN `target_contact` ON `target_contact`.`target_id` = `target`.`id`
                          JOIN `mode` ON `mode`.`id` = `target_contact`.`mode_id`
                          WHERE `target`.`name` = %s AND `mode`.`name` = %s''',
                       (message['target'], target_fallback_mode))
        [(destination, mode, mode_id)] = cursor
        cursor.close()
        connection.close()

        old_mode = message.get('mode', '')
        message['destination'] = destination
        message['mode'] = mode
        message['mode_id'] = mode_id
        update_message_mode(message)
        message_id = message.get('message_id')
        if message_id:
            auditlog.message_change(
                message_id, auditlog.MODE_CHANGE, old_mode, message['mode'],
                'Changing mode due to original mode failure')
        return True
    # target doesn't have email either - bail
    except ValueError:
        logger.exception('target does not have mode(%s) %r', target_fallback_mode, message)
        message['destination'] = message['mode'] = message['mode_id'] = None
        return False


def set_target_contact_by_priority(message):
    session = db.Session()
    result = session.execute('''
              SELECT `target_contact`.`destination` AS dest, `mode`.`name` AS mode_name, `mode`.`id` AS mode_id
              FROM `mode`
              JOIN `target` ON `target`.`name` = :target
              JOIN `target_type` ON `target_type`.`id` = `target`.`type_id`
              JOIN `application` ON `application`.`name` = :application

              -- left join because the "drop" mode isn't going to have a target_contact entry
              LEFT JOIN `target_contact` ON `target_contact`.`mode_id` = `mode`.`id` AND `target_contact`.`target_id` = `target`.`id`

              WHERE mode.id = IFNULL((
                        SELECT `target_application_mode`.`mode_id`
                        FROM `target_application_mode`
                        WHERE `target_application_mode`.`target_id` = target.id AND
                                `target_application_mode`.`application_id` = `application`.`id` AND
                                `target_application_mode`.`priority_id` = :priority_id
                    ), IFNULL(
                      -- 2. Lookup default setting for this app
                      (
                          SELECT `default_application_mode`.`mode_id`
                          FROM `default_application_mode`
                          WHERE    `default_application_mode`.`priority_id` = :priority_id AND
                                `default_application_mode`.`application_id` = `application`.`id`
                      ), IFNULL(
                        -- 3. lookup default user setting
                            (
                                SELECT `target_mode`.`mode_id`
                                FROM `target_mode`
                                WHERE `target_mode`.`target_id` = target.id AND
                                        `target_mode`.`priority_id` = :priority_id
                            ), (
                        -- 4. lookup default iris setting
                                SELECT `mode_id`
                                FROM `priority`
                                WHERE `id` = :priority_id
                            )
                        )
                   )
                )
                -- Make sure this mode is allowed for this application. Eg important apps can't drop.
                AND EXISTS (SELECT 1 FROM `application_mode`
                            WHERE `application_mode`.`mode_id` = `mode`.`id`
                            AND   `application_mode`.`application_id` = `application`.`id`)
                -- And ensure this only works for users
                AND `target_type`.`name` = 'user'
        ''', message)

    try:
        [(destination, mode, mode_id)] = result
    except ValueError:
        raise
    finally:
        session.close()

    if not destination and mode != 'drop':
        logger.error('Did not find destination for message %s and mode is not drop', message)
        return False

    message['destination'] = destination
    message['mode'] = mode
    message['mode_id'] = mode_id

    return True


def set_target_contact(message):
    # If we already have a destination set (eg incident tracking emails or literal_target notifications) no-op this
    if 'destination' in message and not message.get('multi-recipient'):
        return True

    # returns True if contact has been set (even if it has been changed to the fallback). Otherwise, returns False
    destination_query = '''
                SELECT `destination` FROM `target_contact`
                JOIN `target` ON `target`.`id` = `target_contact`.`target_id`
                JOIN `target_type` on `target_type`.`id` = `target`.`type_id`
                WHERE `target`.`name` = %(target)s
                AND `target_type`.`name` = 'user'
                AND (`target_contact`.`mode_id` = %(mode_id)s OR `target_contact`.`mode_id` = (SELECT `id` FROM `mode` WHERE `name` = %(mode)s))
                LIMIT 1'''

    # Handle multi-recipient messages. These are only allowed when mode is specified, not priority.
    # Return True if we're able to resolve at least one target
    if message.get('multi-recipient'):
        connection = db.engine.raw_connection()
        cursor = connection.cursor()
        for t in message['target']:
            try:
                cursor.execute(destination_query, {'target': t['target'], 'mode_id': message.get('mode_id'), 'mode': message.get('mode')})
                if t.get('bcc'):
                    message['bcc_destination'].append(cursor.fetchone()[0])
                else:
                    message['destination'].append(cursor.fetchone()[0])
            except (ValueError, TypeError):
                continue
        cursor.close()
        connection.close()
        return bool(message.get('destination') or message.get('bcc_destination'))

    try:
        if 'mode' in message or 'mode_id' in message:
            # for out of band notification, we already have the mode *OR*
            # mode_id set by API
            connection = db.engine.raw_connection()
            cursor = connection.cursor()
            cursor.execute(destination_query, {'target': message['target'], 'mode_id': message.get('mode_id'), 'mode': message.get('mode')})
            message['destination'] = cursor.fetchone()[0]
            cursor.close()
            connection.close()
            result = True
        elif 'category' in message:
            connection = db.engine.raw_connection()
            cursor = connection.cursor()
            cursor.execute('''
                SELECT `mode`.`id`, `mode`.`name` FROM `category_override`
                JOIN `target` ON `target`.`id` = `category_override`.`user_id`
                JOIN `mode` ON `mode`.`id` = `category_override`.`mode_id`
                WHERE `target`.`name` = %(target)s AND `category_override`.`category_id` = %(category_id)s
            ''', {'target': message['target'], 'category_id': message['category_id']})
            override_mode = cursor.fetchone()
            if override_mode:
                message['mode_id'] = override_mode[0]
                message['mode'] = override_mode[1]
            else:
                message['mode_id'] = message['category_mode_id']
                message['mode'] = message['category_mode']
            cursor.execute('''
                SELECT `destination` FROM `target_contact`
                JOIN `target` ON `target`.`id` = `target_contact`.`target_id`
                JOIN `target_type` on `target_type`.`id` = `target`.`type_id`
                WHERE `target`.`name` = %(target)s
                AND `target_type`.`name` = 'user'
                AND `target_contact`.`mode_id` = %(mode_id)s
                LIMIT 1
                ''', {'target': message['target'], 'mode_id': message['mode_id']})
            message['destination'] = cursor.fetchone()[0]
            cursor.close()
            connection.close()
            result = True
        else:
            # message triggered by incident will only have priority
            result = set_target_contact_by_priority(message)
        if result:
            cache.target_reprioritization(message)
        else:
            logger.warning('target does not have mode %r', message)
            result = set_target_fallback_mode(message)
        return result
    except (ValueError, TypeError):
        logger.exception('target does not have mode %r', message)
        return set_target_fallback_mode(message)


def render(message):
    if not message.get('template'):
        if message.get('message_id'):
            # email response from iris does not use template this means the
            # message content is already in DB
            connection = db.engine.raw_connection()
            cursor = connection.cursor()
            cursor.execute('SELECT `body`, `subject` FROM `message` WHERE `id` = %s',
                           message['message_id'])
            msg_content = cursor.fetchone()
            message['body'], message['subject'] = msg_content[0], msg_content[1]
            cursor.close()
            connection.close()
        else:
            # out of band message does not have id and should already have the
            # content populated
            return
    elif 'aggregated_ids' in message:
        message['subject'] = '[%%(application)s] %s messages from plan %%(plan)s' % len(message['aggregated_ids']) % message
        message['body'] = 'Batch ID: %(batch_id)s' % message
        message['template_id'] = None
    else:
        if message.get('body') is None:
            message['body'] = ''
        error = None
        try:
            template = cache.templates[message['template']]
            try:
                application_template = template[message['application']]
                try:
                    # When we want to "render" a dropped message, treat it as if it's an email
                    mode_template = application_template['email' if message['mode'] == 'drop' else message['mode']]
                    try:
                        message['subject'] = mode_template['subject'].render(**message['context'])
                    except Exception as e:
                        error = 'template %(template)s - %(application)s - %(mode)s - subject failed to render: ' + str(e)
                    try:
                        message['body'] += mode_template['body'].render(**message['context'])
                    except Exception as e:
                        error = 'template %(template)s - %(application)s - %(mode)s - body failed to render: ' + str(e)
                    message['template_id'] = template['id']
                except KeyError:
                    error = 'template %(template)s - %(application)s does not have mode %(mode)s'
            except KeyError:
                error = 'template %(template)s does not have application %(application)s'
        except KeyError:
            error = 'template %(template)s does not exist'

        if error:
            # TODO:
            # define application default template here
            # additionally, instead of repr() as the final message render
            # format define a generic template that will work for all contexts
            # - even those that are invalid as a final final format maybe repr or pprint
            logger.error(error, message)
            message['subject'] = '%(message_id)s Iris failed to render your message' % message
            message['body'] = 'Failed rendering message.\n\nContext: %s\n\nError: %s' % (repr(message), error % message)
            message['template_id'] = None
        else:
            if config.get('enable_gmail_oneclick') and message['mode'] == 'email' and 'incident_id' in message:
                oneclick_url = generate_oneclick_url(config, {
                    'msg_id': message['message_id'],
                    'email_address': message['destination'],
                    'cmd': 'claim'
                })
                additional_body = oneclick_email_markup % {
                    'url': oneclick_url,
                    'incident_id': message['incident_id']
                }
                message['extra_html'] = additional_body


def mark_message_as_sent(message):
    connection = db.engine.raw_connection()

    params = [
        message['destination'],
        message['mode_id'],
        message.get('template_id'),
    ]

    if 'aggregated_ids' in message:
        sql = SENT_MESSAGE_BATCH_SQL % connection.escape(message['aggregated_ids'])
        params.append(message['batch_id'])
        message_ids = message['aggregated_ids']
    else:
        sql = SENT_MESSAGE_SQL
        params.append(message['message_id'])
        message_ids = [message['message_id']]

    cursor = connection.cursor()
    if not message['subject']:
        message['subject'] = ''
        logger.warning('Message id %s has blank subject', message.get('message_id', '?'))

    max_retries = 3

    # this deadlocks sometimes. try until it doesn't.
    for i in range(1, max_retries + 1):
        try:
            cursor.execute(sql, params)
            connection.commit()
            break
        except DataError:
            logger.warning('Failed updating message metadata status (message ID %s) (application %s)', message.get('message_id', '?'), message.get('application', '?'))
            break
        except Exception:
            if i == max_retries:
                logger.warning('Failed running sent message update query. (Try %s/%s)', i, max_retries)
            else:
                logger.warning('Failed running sent message update query. (Try %s/%s)', i, max_retries)
            sleep(.2)

    # Clean messages cache
    for message_id in message_ids:
        messages.pop(message_id, None)

    # Update subject and body separately, as they may fail and we don't necessarily care if they do
    if len(message['subject']) > 255:
        message['subject'] = message['subject'][:255]

    if len(message['body']) > MAX_MESSAGE_BODY_LENGTH:
        logger.warning('Message id %s has a ridiculously long body (%s chars). Truncating it.',
                       message.get('message_id', '?'), len(message['body']))
        message['body'] = message['body'][:MAX_MESSAGE_BODY_LENGTH]

    if 'aggregated_ids' in message:
        update_ids = tuple(message['aggregated_ids'])
    else:
        update_ids = tuple([message['message_id']])

    max_retries = 3

    # this deadlocks sometimes. try until it doesn't.
    for i in range(1, max_retries + 1):
        try:
            cursor.execute(UPDATE_MESSAGE_BODY_SQL, (message['body'], message['subject'], update_ids))
            connection.commit()
            break
        except DataError:
            logger.warning('Failed updating message body+subject (message IDs %s) (application %s)', update_ids, message.get('application', '?'))
            break
        except Exception:
            if i == max_retries:
                logger.warning('Failed updating message body+subject (message IDs %s) (application %s) (Try %s/%s)', update_ids, message.get('application', '?'), i, max_retries)
            else:
                logger.warning('Failed updating message body+subject (message IDs %s) (application %s) (Try %s/%s)', update_ids, message.get('application', '?'), i, max_retries)
            sleep(.2)

    cursor.close()
    connection.close()


def update_message_sent_status(message, status):
    message_id = message.get('message_id')
    if not message_id:
        return

    mode = message.get('mode')

    if not mode:
        return

    # Don't track this for twilio as those are kept track of separately. Make
    # use of this for email, and, as a side effect of that for slack
    if mode in ('sms', 'call'):
        return

    session = db.Session()
    retries = 0
    max_retries = 3
    while True:
        retries += 1
        try:
            session.execute('''INSERT INTO `generic_message_sent_status` (`message_id`, `status`)
                        VALUES (:message_id, :status)
                        ON DUPLICATE KEY UPDATE `status` =  :status''',
                            {'message_id': message_id, 'status': status})
            session.commit()
        except Exception:
            logger.warning('Failed setting message sent status for message %s (Try %s/%s)', message_id, retries, max_retries)
            if retries < max_retries:
                sleep(.2)
                continue
            else:
                raise Exception('Failed setting message sent status for message retries exceeded')
        else:
            break

    session.close()


def mark_message_has_no_contact(message):
    message_id = message.get('message_id')
    if not message_id:
        logger.warning('Cannot mark message "%s" as not having contact as message_id is missing', message)
        return

    connection = db.engine.raw_connection()
    cursor = connection.cursor()
    retries = 0
    max_retries = 3
    while True:
        retries += 1
        try:
            cursor.execute('UPDATE `message` set `active`=0 WHERE `id`=%s',
                           message_id)
            connection.commit()
            cursor.close()
        except Exception:
            logger.warning('Failed setting message %s as not having contact (Try %s/%s)', message_id, retries, max_retries)
            if retries < max_retries:
                sleep(.2)
                continue
            else:
                raise Exception('Failed setting message as not having contact')
        else:
            break
    connection.close()
    auditlog.message_change(
        message_id, auditlog.MODE_CHANGE, target_fallback_mode, 'invalid',
        'Ignore message as we failed to resolve target contact')


def distributed_send_message(message, vendor_manager):
    # If I am the master and this message isn't for a slave, attempt
    # sending my messages through my slaves.
    if not message.get('to_slave') and coordinator.am_i_master():
        try:
            if coordinator.slave_count and coordinator.slaves:
                for i, address in enumerate(coordinator.slaves):
                    if i >= coordinator.slave_count:
                        logger.error('Failed using all configured slaves; resorting to local send_message')
                        break
                    if rpc.send_message_to_slave(message, address):
                        return True, False
        except StopIteration:
            logger.warning('No more slaves. Sending locally.')

    logger.info('Sending message (ID %s) locally', message.get('message_id', '?'))

    runtime = vendor_manager.send_message(message)
    add_mode_stat(message['mode'], runtime)

    # application is not present for incident tracking emails
    if 'application' in message:
        notification_count = 1 if not message.get('target_list') else len(message['destination'])
        metrics_key = 'app_%(application)s_mode_%(mode)s_cnt' % message
        metrics.add_new_metrics({metrics_key: 0})
        metrics.incr(metrics_key, inc=notification_count)

    if runtime is not None:
        return True, True

    return False, False


def fetch_and_send_message(send_queue, vendor_manager):
    try:
        message = send_queue.get(True, 4)
    except queue.Empty:
        return

    metrics.incr('send_queue_gets_cnt')

    retry_count = message.get('retry_count')
    is_retry = retry_count is not None
    if is_retry and retry_count >= MAX_MESSAGE_RETRIES:
        logger.warning('Maximum retry count for app: %s target:%s breached', message.get('application', '?'), message.get('target', '?'))
        return

    if not is_retry:
        sanitize_unicode_dict(message)

    if 'message_id' not in message:
        message['message_id'] = None

    message_to_slave = message.get('to_slave', False)

    drop_mode_id = api_cache.modes.get('drop')

    # If this app breaches hard quota, drop message on floor, and update in UI if it has an ID
    if not is_retry and not message_to_slave and not quota.allow_send(message):
        logger.warning('Hard message quota exceeded; Dropping this message on floor: %s', message)
        if message['message_id']:
            spawn(auditlog.message_change,
                  message['message_id'], auditlog.MODE_CHANGE, message.get('mode', '?'), 'drop',
                  'Dropping due to hard quota violation.')

            # If we know the ID for the mode drop, reflect that for the message
            if drop_mode_id:
                message['mode'] = 'drop'
                message['mode_id'] = drop_mode_id
            else:
                logger.error('Can\'t mark message %s as dropped as we don\'t know the mode ID for %s',
                             message, 'drop')

            # Render, so we're able to populate the message table with the
            # proper subject/etc as well as information that it was dropped.
            render(message)
            mark_message_as_sent(message)
        return

    # If we're set to drop this message, no-op this before message gets sent to a vendor
    if message.get('mode') == 'drop':
        if message['message_id']:
            render(message)
            mark_message_as_sent(message)
        add_mode_stat('drop', 0)

        metrics_key = 'app_%(application)s_mode_drop_cnt' % message
        metrics.add_new_metrics({metrics_key: 0})
        metrics.incr(metrics_key)

        return

    # Only render this message and validate its body/etc if it's not a retry, in which case this
    # step would have been done before
    if not is_retry and not message_to_slave:
        render(message)

        if message.get('body') is None:
            message['body'] = ''

        # Drop this message, and mark it as dropped, rather than sending it, if its
        # body is too long and we were normally going to send it anyway.
        body_length = len(message['body'])
        if body_length > MAX_MESSAGE_BODY_LENGTH:
            logger.warning('Message id %s has a ridiculously long body (%s chars). Dropping it.',
                           message['message_id'], body_length)
            spawn(auditlog.message_change,
                  message['message_id'], auditlog.MODE_CHANGE, message.get('mode', '?'), 'drop',
                  'Dropping due to excessive body length (%s > %s chars)' % (
                      body_length, MAX_MESSAGE_BODY_LENGTH))

            metrics.incr('msg_drop_length_cnt')

            # Truncate this here to avoid a duplicate log message in
            # mark_message_as_sent(), as we still need to call that to update the body/subject
            message['body'] = message['body'][:MAX_MESSAGE_BODY_LENGTH]

            if drop_mode_id:
                message['mode'] = 'drop'
                message['mode_id'] = drop_mode_id

            mark_message_as_sent(message)
            return

    success = False
    sent_locally = False
    try:
        success, sent_locally = distributed_send_message(message, vendor_manager)
    except Exception:
        logger.warning('Failed to send message: %s', message)
        add_mode_stat(message['mode'], None)
    if not success and not sent_locally:
        if not message.get('unexpanded') and message['mode'] != 'email':
            if message.get('target'):
                logger.error('reclassifying as email %s', message)
                set_target_fallback_mode(message)
                render(message)
                try:
                    success, sent_locally = distributed_send_message(message, vendor_manager)
                # nope - log and bail
                except Exception:
                    metrics.incr('task_failure')
                    add_mode_stat(message['mode'], None)
                    logger.error('unable to send %s', message)
                    sent_locally = True

    # Take it out of our list of active queued messages if it's there
    if message['message_id']:
        message_ids_being_sent.discard(message['message_id'])

    if success:
        metrics.incr('message_send_cnt')
        if message['message_id'] and sent_locally:
            mark_message_as_sent(message)

        if message_to_slave:
            metrics.incr('slave_message_send_success_cnt')

    else:
        # If we're not successful, try retrying it
        message['retry_count'] = message.get('retry_count', 0) + 1
        message_send_enqueue(message)
        metrics.incr('message_retry_cnt')
        logger.info('Message %s failed. Re-queuing for retry (%s/%s).', message, message['retry_count'], MAX_MESSAGE_RETRIES)

        if message_to_slave:
            metrics.incr('slave_message_send_fail_cnt')

    if message['message_id'] and sent_locally:
        update_message_sent_status(message, success)


def worker(send_queue, worker_config, kill_set):
    vendor_manager = IrisVendorManager(worker_config.get('vendors', []), worker_config.get('applications', []))
    while True:
        fetch_and_send_message(send_queue, vendor_manager)

        if kill_set.is_set():
            vendor_manager.cleanup()
            return


def maintain_workers(config):
    # We mangle this config dict a bit and pass it around. Avoid latering the main one
    config = copy.deepcopy(config)

    # Determine counts for "normal" modes
    default_worker_count = 100

    try:
        worker_count = int(config.get('sender_workers', default_worker_count))
    except ValueError:
        worker_count = default_worker_count

    workers_per_mode_cnt = worker_count // len(api_cache.modes)
    workers_per_mode = {mode: workers_per_mode_cnt for mode in api_cache.modes}

    # Email scale up/down. For this trickyness to work: 1) Only one SMTP vendor 2) It has "autoscale" enabled
    # 3) It uses SMTP gateway instead of hard coded mx_servers

    email_vendor = None
    email_mx_gateway = None
    email_smtp_workers = None
    configured_vendors = config['vendors']
    for vendor in configured_vendors:
        if vendor['type'] == 'iris_smtp':
            email_mx_gateway = vendor.get('smtp_gateway')
            if not email_mx_gateway:
                break

            if not vendor.get('autoscale'):
                logger.warning('Ignoring possibly auto scaled MX gateway %s', email_mx_gateway)
                break

            email_vendor = vendor
            break

    # Hard code workers for each mode manually in config
    workers_per_mode_config = config.get('sender', {}).get('workers_per_mode', {})
    workers_per_mode.update({mode: int(workers_per_mode_config[mode]) for mode in api_cache.modes if mode in workers_per_mode_config})

    email_autoscale = email_vendor is not None

    if email_autoscale:
        try:
            email_smtp_workers = iris_smtp.iris_smtp.determine_worker_count(email_vendor)
        except Exception:
            logger.warning('Failed determining MX records this round')
            email_smtp_workers = None

        old_email_worker_count = workers_per_mode.pop('email', None)
        if old_email_worker_count:
            logger.info('Going with autoscaled smtp workers instead of %d', old_email_worker_count)

        # Remove all smtp vendor objects so they don't get initialized unnecessarily
        config['vendors'] = [vendor for vendor in config['vendors'] if vendor['type'] != 'iris_smtp']

    logger.info('Workers per mode: %s', ', '.join('%s: %s' % count for count in iter(workers_per_mode.items())))

    # Make sure all the counts and distributions for "normal" workers are proper, including email if we're not doing MX record
    # autoscaling

    for mode, worker_count in workers_per_mode.items():
        mode_tasks = worker_tasks[mode]
        if mode_tasks:
            for task in mode_tasks:
                if not bool(task['greenlet']):
                    logger.error("worker task for mode %s failed, %s. Respawning", mode, task['greenlet'].exception)
                    kill_set = gevent.event.Event()
                    task.update({'greenlet': spawn(worker, per_mode_send_queues[mode], config, kill_set), 'kill_set': kill_set})
                    metrics.incr('workers_respawn_cnt')
        else:
            for x in range(worker_count):
                kill_set = gevent.event.Event()
                mode_tasks.append({'greenlet': spawn(worker, per_mode_send_queues[mode], config, kill_set), 'kill_set': kill_set})

    # Maintain, grow, and shrink email workers
    if email_autoscale and email_smtp_workers:
        logger.info('Configuring auto scaling smtp records')
        email_queue = per_mode_send_queues['email']
        tasks_to_kill = []

        # Adjust worker count
        for mx, correct_worker_count in email_smtp_workers.items():
            mx_workers = autoscale_email_worker_tasks[mx]
            current_task_count = len(mx_workers)

            # Correct amount of workers
            if current_task_count == correct_worker_count:
                logger.info('MX record %s has the correct number of tasks %d', mx, correct_worker_count)

            # Need more workers
            elif current_task_count < correct_worker_count:
                new_task_count = correct_worker_count - current_task_count
                logger.info('Auto scaling MX record %s UP %d tasks', mx, new_task_count)

                # Configure this email worker with just one hard coded mx server, the one this
                # worker will correspond to
                email_vendor_config = copy.deepcopy(email_vendor)
                email_vendor_config['smtp_server'] = mx
                email_vendor_config.pop('smtp_gateway', None)
                worker_config = copy.deepcopy(config)
                worker_config['vendors'] = [email_vendor_config]

                for x in range(new_task_count):
                    kill_set = gevent.event.Event()
                    mx_workers.append({'greenlet': spawn(worker, email_queue, worker_config, kill_set), 'kill_set': kill_set})

            # Need less workers
            elif current_task_count > correct_worker_count:
                kill_task_count = current_task_count - correct_worker_count
                logger.info('Auto scaling MX record %s DOWN %d tasks', mx, kill_task_count)
                for x in range(kill_task_count):
                    try:
                        tasks_to_kill.append(mx_workers.pop())
                    except IndexError:
                        break

        # Kill MX records no longer in use
        kill_mx = autoscale_email_worker_tasks.keys() - email_smtp_workers.keys()
        for mx in kill_mx:
            workers = autoscale_email_worker_tasks[mx]
            if workers:
                logger.info('Removing %d tasks for unused MX %s', len(workers), mx)
                tasks_to_kill += workers
            del autoscale_email_worker_tasks[mx]

        # Make sure all existing workers are alive
        for mx, mx_tasks in autoscale_email_worker_tasks.items():
            email_vendor_config = copy.deepcopy(email_vendor)
            email_vendor_config['smtp_server'] = mx
            email_vendor_config.pop('smtp_gateway', None)
            worker_config = copy.deepcopy(config)
            worker_config['vendors'] = [email_vendor_config]

            for task in mx_tasks:
                if not bool(task['greenlet']):
                    kill_set = gevent.event.Event()
                    task.update({'greenlet': spawn(worker, email_queue, worker_config, kill_set), 'kill_set': kill_set})
                    logger.error("worker email task for mx %s failed, %s. Respawning", mx, task['greenlet'].exception)
                    metrics.incr('workers_respawn_cnt')

        # Kill all of the workers who shouldn't exist anymore
        for task in tasks_to_kill:
            task['kill_set'].set()


def gwatch_renewer():
    gmail_config = config['gmail']
    gcli = Gmail(gmail_config, config.get('gmail_proxy'))
    while True:
        # If we stop being master, bail out of this
        if coordinator is not None and not coordinator.am_i_master():
            return

        logger.info('[-] start gmail watcher loop...')
        logger.info('renewing gmail watcher...')
        re = gcli.watch(gmail_config['project'], gmail_config['topic'])
        try:
            history_id, expiration = (int(re['historyId']),
                                      int(re['expiration']) / 1000 - time.time())
        except KeyError:
            logger.exception('[*] gmail watcher run failed. Skipping this run.')
        else:
            metrics.set('gmail_history_id', history_id)
            metrics.set('gmail_seconds_to_watch_expiration', expiration)
            logger.info('[*] gmail watcher loop finished')

        # only renew every 8 hours
        sleep(60 * 60 * 8)


def prune_old_audit_logs_worker():
    while True:
        # If we stop being master, bail out of this
        if coordinator is not None and not coordinator.am_i_master():
            return

        try:
            connection = db.engine.raw_connection()
            cursor = connection.cursor()
            cursor.execute(PRUNE_OLD_AUDIT_LOGS_SQL)
            connection.commit()
            cursor.close()
        except Exception:
            logger.warning('Failed pruning old audit logs')
        finally:
            connection.close()

        logger.info('Ran task to prune old audit logs. Waiting 4 hours until next run.')
        sleep(60 * 60 * 4)


def mock_gwatch_renewer():
    while True:
        logger.info('[-] start mock gmail watcher loop...')
        logger.info('[*] mock gmail watcher loop finished')
        sleep(60)


def sender_shutdown():
    global shutdown_started

    # Make control+c or some other kill signal force quit the second time it happens
    if shutdown_started:
        logger.info('Force exiting')
        os._exit(0)
    else:
        shutdown_started = True
        logger.info('Shutting server..')

    # Immediately release all locks and give up any master status and slave presence
    if coordinator:
        coordinator.leave_cluster()

    # Stop sender RPC server
    rpc.shutdown()

    for tasks in worker_tasks.values():
        for task in tasks:
            task['kill_set'].set()

    for tasks in autoscale_email_worker_tasks.values():
        for task in tasks:
            task['kill_set'].set()

    logger.info('Waiting for sender workers to shut down')

    for tasks in worker_tasks.values():
        for task in tasks:
            task['greenlet'].join()

    for tasks in autoscale_email_worker_tasks.values():
        for task in tasks:
            task['greenlet'].join()

    # Force quit. Avoid sender process existing longer than it needs to
    os._exit(0)


def modify_restricted_calls(message):
    # check for known corner cases in message delivery and correct accordingly

    # due to calling restrictions to china override mode and send as sms instead
    if message['destination'].startswith('+86'):
        message['mode'] = 'sms'
        auditlog.message_change(
            message['message_id'], auditlog.MODE_CHANGE, 'call', 'sms',
            'Changing mode due to calling restriction')
        # this message will appear before the contents of the sms template
        message['body'] = 'Due to legal restrictions we are unable to deliver calls to China. Please check Iris for complete incident details and change your settings to use sms, slack, or email instead - ' + message.get('body', '')


def modify_restricted_sms(message):

    if message.get('incident_id'):
        # Alter message to get around SMS restrictions in countries like India
        # Replace body with pre-registered template so it won't be dropped by carriers

        destination = message['destination']
        message_id = message['message_id']
        mode_id = message['mode_id']
        connection = db.engine.raw_connection()
        cursor = connection.cursor()
        # only replace messages for users that have opted in
        cursor.execute('''SELECT count(`mode_template_override`.`target_id`)
            FROM `mode_template_override` JOIN `target_contact`
            ON `mode_template_override`.`target_id` = `target_contact`.`target_id`
            WHERE `target_contact`.`destination` = %s and `target_contact`.`mode_id` = %s''', (destination, mode_id))
        result = cursor.fetchone()[0]
        connection.close()
        if result:
            body_template = config.get('sms_override_template')
            message['body'] = body_template % message.get('incident_id')
            auditlog.message_change(message_id, auditlog.CONTENT_CHANGE, 'sms', 'sms',
                                    'Replaced SMS body to avoid being blocked by carrier')


def message_send_enqueue(message):

    # If this message has an ID, avoid queueing it twice, assuming it's not a retry
    message_id = message.get('message_id')
    if message_id is not None and message_id in message_ids_being_sent and not message.get('retry_count'):
        logger.debug('Not-requeueing message %s', message)
        return

    # Set the target contact here so we determine the mode, which determines the
    # queue it gets inserted into. Skip this step if it's a retry because we'd
    # already have the message's contact info decided
    if not message.get('retry_count'):
        has_contact = set_target_contact(message)
        if not has_contact:
            mark_message_has_no_contact(message)
            metrics.incr('task_failure')
            logger.error('Failed to send message, no contact found: %s', message)
            return

    message_mode = message.get('mode')
    if message_mode and message_mode in per_mode_send_queues:
        # check for known corner case limitations and correct accordingly
        if message['mode'] == 'call':
            modify_restricted_calls(message)

        if message['mode'] == 'sms':
            modify_restricted_sms(message)

        if message_id is not None:
            message_ids_being_sent.add(message_id)
        per_mode_send_queues[message_mode].put(message)
        metrics.incr('send_queue_puts_cnt')
    else:
        logger.error('Message %s does not have proper mode %s', message, message_mode)
        metrics.incr('send_queue_puts_fail_cnt')


def update_api_cache_worker():
    while True:
        logger.debug('Reinitializing cache')
        api_cache.cache_priorities()
        api_cache.cache_applications()
        api_cache.cache_modes()
        sleep(60)


def log_sender_master():
    while True:
        sender_master = coordinator.get_current_master()
        if sender_master:
            logger.info("Current sender master: %s" % sender_master)
        sleep(30)


def init_sender(config):
    gevent.signal(signal.SIGINT, sender_shutdown)
    gevent.signal(signal.SIGTERM, sender_shutdown)
    gevent.signal(signal.SIGQUIT, sender_shutdown)

    process_title = config['sender'].get('process_title')

    if process_title and isinstance(process_title, str):
        setproctitle.setproctitle(process_title)
        logger.info('Changing process name to %s', process_title)

    api_host = config['sender'].get('api_host', 'http://localhost:16649')
    db.init(config)
    cache.init(api_host, config)
    metrics.init(config, 'iris-sender', default_sender_metrics)
    api_cache.cache_priorities()
    api_cache.cache_applications()
    api_cache.cache_modes()

    global should_mock_gwatch_renewer, send_message
    if config['sender'].get('debug'):
        logger.info('DEBUG MODE')
        should_mock_gwatch_renewer = True
        should_skip_send = True
    else:
        should_skip_send = False
    should_mock_gwatch_renewer = should_mock_gwatch_renewer or config.get('skipgmailwatch', False)
    should_skip_send = should_skip_send or config.get('skipsend', False)

    if should_skip_send:
        config['vendors'] = [{
            'type': 'iris_dummy',
            'name': 'iris dummy vendor'
        }]

    global quota
    quota = ApplicationQuota(db, cache.targets_for_role, message_send_enqueue, config['sender'].get('sender_app'))

    global coordinator
    zk_hosts = config['sender'].get('zookeeper_cluster', False)

    if zk_hosts:
        logger.info('Initializing coordinator with ZK: %s', zk_hosts)
        from iris.coordinator.kazoo import Coordinator
        coordinator = Coordinator(zk_hosts=zk_hosts,
                                  hostname=socket.gethostname(),
                                  port=config['sender'].get('port', 2321),
                                  join_cluster=True)
    else:
        logger.info('ZK cluster info not specified. Using master status from config')
        from iris.coordinator.noncluster import Coordinator
        coordinator = Coordinator(is_master=config['sender'].get('is_master', True),
                                  slaves=config['sender'].get('slaves', []))


def main():
    global config
    global shutdown_started
    config = load_config()

    start_time = time.time()

    logger.info('[-] bootstraping sender...')
    init_sender(config)
    spawn(update_api_cache_worker)
    init_plugins(config.get('plugins', {}))

    if not rpc.run(config['sender']):
        sender_shutdown()

    for mode in api_cache.modes:
        per_mode_send_queues[mode] = queue.Queue()

    rpc.init(config['sender'], dict(
        message_send_enqueue=message_send_enqueue
    ))

    spawn(coordinator.update_forever)
    spawn(log_sender_master)
    send_task = spawn(send)

    maintain_workers(config)

    disable_gwatch_renewer = config['sender'].get('disable_gwatch_renewer', False)
    gwatch_renewer_task = None
    prune_audit_logs_task = None

    interval = 60
    logger.info('[*] sender bootstrapped')
    while True:

        # When the shutdown starts, avoid doing sender tasks but keep this
        # loop open as the shutdown function terminates the app once messages
        # are done sending.
        if shutdown_started:
            logger.info('--> Shutdown in progress')
            sleep(30)
            continue

        runtime = int(time.time())
        logger.info('--> sender looop started.')

        cache.refresh()
        cache.purge()

        # If we're currently a master, ensure our master-greenlets are running
        # and we're doing the master duties
        if coordinator.am_i_master():
            if not disable_gwatch_renewer and not bool(gwatch_renewer_task):
                if should_mock_gwatch_renewer:
                    gwatch_renewer_task = spawn(mock_gwatch_renewer)
                else:
                    gwatch_renewer_task = spawn(gwatch_renewer)

            if not bool(prune_audit_logs_task):
                prune_audit_logs_task = spawn(prune_old_audit_logs_worker)

            try:
                escalate()
                deactivate()
                poll()
                aggregate(runtime)
            except Exception:
                metrics.incr('task_failure')
                logger.exception("Exception occured in main loop.")

        # If we're not master, don't do the master tasks and make sure those other
        # greenlets are stopped if they're running
        else:
            logger.info('I am not the master so I am not doing master sender tasks.')

            # Stop these task greenlets if they're running. Technically this should
            # never happen because if we're the master, we'll likely only stop being the
            # master if our process exits, which would kill these greenlets anyway.
            if bool(gwatch_renewer_task):
                logger.info('I am not master anymore so stopping the gwatch renewer')
                gwatch_renewer_task.kill()

            if bool(prune_audit_logs_task):
                logger.info('I am not master anymore so stopping the audit logs worker')
                prune_audit_logs_task.kill()

        # check status for all background greenlets and respawn if necessary
        if not bool(send_task):
            logger.error("send task failed, %s", send_task.exception)
            metrics.incr('task_failure')
            send_task = spawn(send)

        for mode, send_queue in per_mode_send_queues.items():

            # Set metric for size of worker queue
            metrics.set('send_queue_%s_size' % mode, len(send_queue))

        maintain_workers(config)

        now = time.time()
        metrics.set('sender_uptime', int(now - start_time))

        metrics.set('message_ids_being_sent_cnt', len(message_ids_being_sent))

        spawn(metrics.emit)

        elapsed_time = now - runtime
        nap_time = max(0, interval - elapsed_time)

        logger.info('--> sender loop finished in %s seconds - sleeping %s seconds',
                    elapsed_time, nap_time)
        sleep(nap_time)


if __name__ == '__main__':
    main()



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# coding: utf-8
# pylint: disable=star-args

from base64 import urlsafe_b64encode
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
# from getpass import getuser
from logging import getLogger
from mimetypes import guess_type
from os import makedirs
from os.path import basename, exists, join

from googleapiclient import errors
from googleapiclient.discovery import build
from httplib2 import Http, ProxyInfo, socks
from oauth2client import client, tools
from oauth2client.file import Storage
# TODO: figure out why keyring fails on OS X
# from oauth2client.keyring_storage import Storage

logger = getLogger(__name__)


class Gmail(object):
    """
    :param config: gmail configuration
    """
    def __init__(self, config, proxy_cfg=None):
        self.config = config
        if proxy_cfg:
            proxy_info = ProxyInfo(
                socks.PROXY_TYPE_HTTP_NO_TUNNEL,
                proxy_cfg['host'],
                proxy_cfg['port']
            )
        else:
            proxy_info = None
        self.http = Http(proxy_info=proxy_info)
        self.client = None

    def _get_credentials(self):
        """Get OAuth credentials

        :return: OAuth credentials
        :rtype: :class:`oauth2client.client.Credentials`
        """
        credential_dir = join(self.config['creds_cache_dir'], 'cached_oauth_credentials')
        if not exists(credential_dir):
            makedirs(credential_dir)
        credential_path = join(credential_dir, 'googleapis.json')

        store = Storage(credential_path)
        credentials = store.get()
        if not credentials or credentials.invalid:
            flow = client.flow_from_clientsecrets(self.config['creds'],
                                                  self.config['scope'])
            flow.user_agent = 'Iris Gmail Integration'
            credentials = tools.run_flow(
                flow, store, tools.argparser.parse_args(args=['--noauth_local_webserver']))
            logger.info('Storing credentials to %s' % credential_path)
        else:
            credentials.refresh(self.http)
        return credentials

    def _get_http(self):
        """Construct httplib2.Http resource.

        :return: An object through which HTTP request will be made.
        :rtype: :class:`httplib2.Http`
        """
        return self._get_credentials().authorize(self.http)

    def connect(self):
        """Construct a Resource for interacting with the Gmail v1 API.

        :rtype: `None`
        """
        if self.client is None:
            self.client = build('gmail', 'v1', http=self._get_http())

    @staticmethod
    def create_message(
            sender,
            receiver,
            subject,
            payload):
        """Create a message for an email.

        :param sender: Email address of the sender.
        :param receiver: Email address of the receiver.
        :param subject: The subject of the email message.
        :param payload: The text of the email message.
        :return: An object containing a base64url encoded email object.
        :rtype: dict
        """
        message = MIMEText(payload)
        message['to'] = receiver
        message['from'] = sender
        message['subject'] = subject
        return {'raw': urlsafe_b64encode(message.as_string())}

    @staticmethod
    def create_message_with_attachment(
            sender,
            receiver,
            subject,
            message_text,
            path):
        """Create a message for an email.

        :param sender: Email address of the sender.
        :param receiver: Email address of the receiver.
        :param subject: The subject of the email message.
        :param message_text: The text of the email message.
        :param path: The name of the file to be attached.
        :return: An object containing a base64url encoded email object.
        :rtype: dict
        """
        message = MIMEMultipart()
        message['to'] = receiver
        message['from'] = sender
        message['subject'] = subject

        msg = MIMEText(message_text)
        message.attach(msg)

        content_type, encoding = guess_type(path)

        if content_type is None or encoding is not None:
            content_type = 'application/octet-stream'

        main_type, sub_type = content_type.split('/', 1)

        with open(path, 'r') as attachment:
            if main_type == 'text':
                msg = MIMEText(attachment.read(), _subtype=sub_type)
            elif main_type == 'image':
                msg = MIMEImage(attachment.read(), _subtype=sub_type)
            elif main_type == 'audio':
                msg = MIMEAudio(attachment.read(), _subtype=sub_type)
            else:
                msg = MIMEBase(main_type, sub_type)
                msg.set_payload(attachment.read())

        msg.add_header(
            'Content-Disposition',
            'attachment',
            filename=basename(path))
        message.attach(msg)

        return {'raw': urlsafe_b64encode(message.as_string())}

    def send_message(
            self,
            body='',
            user_id='me',
            **kwargs):
        """Send an email message.

        :param body: Message to be sent.
        :param user_id: User's email address. The special value "me" can be used
            to indicate the authenticated user.
        :return: Sent Message.
        :rtype: dict
        """
        self.connect()
        if not body:
            kwargs = dict(list({
                'sender': self.config.get('sub')
            }.items()) + list(kwargs.items()))
            body = self.create_message(**kwargs)
        ret = {}
        try:
            ret = (self.client.users().messages().send(
                userId=user_id,
                body=body
            ).execute())
        except errors.HttpError as error:
            logger.error('An error occurred: %s', error)
        else:
            logger.info('Message Id: %s', ret.get('id', ''))
        return ret

    @staticmethod
    def _fqrn(
            resource_type,
            project,
            resource):
        """Return a fully qualified resource name for Cloud Pub/Sub.

        :param resource_type:
        :param project:
        :param resource:
        :rtype: str
        """
        return "projects/{0}/{1}/{2}".format(project, resource_type, resource)

    def _get_full_topic_name(
            self,
            project,
            topic):
        """Return a fully qualified topic name.

        :param project:
        :param topic:
        :rtype: str
        """
        return self._fqrn('topics', project, topic)

    def watch(
            self,
            project,
            topic,
            user_id='me'):
        """Setup watching on user's Inbox

        :param project:
        :param topic:
        :param user_id: User's email address. The special value "me" can be used
            to indicate the authenticated user.
        :rtype: dict
        """
        self.connect()
        body = {
            'labelIds': ['INBOX', 'UNREAD'],
            'topicName': self._get_full_topic_name(project, topic)
        }
        ret = {}
        try:
            ret = self.client.users().watch(
                userId=user_id,
                body=body
            ).execute()
        except errors.HttpError as error:
            logger.error('An error occurred: %s', error)
        else:
            logger.debug('Watch expiration: %s', ret.get('expiration', ''))
        return ret



#!/usr/bin/env python

# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import gc
import logging
import sys
import multiprocessing
import gunicorn.app.base
import iris
import iris.config
import imp


class StandaloneApplication(gunicorn.app.base.BaseApplication):

    def __init__(self, options=None, skip_build_assets=False):
        self.options = options or {}
        self.skip_build_assets = skip_build_assets
        super(StandaloneApplication, self).__init__()

    def load_config(self):
        config = {key: value for key, value in self.options.items()
                  if key in self.cfg.settings and value is not None}
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        import iris
        imp.reload(iris)
        imp.reload(iris.config)
        config = iris.config.load_config(sys.argv[1])

        import iris.api
        app = iris.api.get_api(config)

        if not self.skip_build_assets:
            for r in gc.get_referrers(self):
                if isinstance(r, dict) and '_num_workers' in r:
                    gunicorn_arbiter = r

            # only build assets on one worker to avoid race conditions
            if gunicorn_arbiter['worker_age'] % self.options['workers'] == 0:
                import iris.ui
                iris.ui.build_assets()

        return app


def main():
    if len(sys.argv) <= 1:
        sys.exit('USAGE: %s CONFIG_FILE [--skip-build-assets]' % sys.argv[0])
    elif len(sys.argv) >= 3:
        skip_build_assets = (sys.argv[2] == '--skip-build-assets')
    else:
        skip_build_assets = False

    logging.basicConfig(format='[%(asctime)s] [%(process)d] [%(levelname)s] %(name)s %(message)s',
                        level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S %z')

    config = iris.config.load_config(sys.argv[1])
    server = config['server']

    options = {
        'preload_app': False,
        'reload': True,
        'bind': '%s:%s' % (server['host'], server['port']),
        'worker_class': 'gevent',
        'accesslog': '-',
        'workers': multiprocessing.cpu_count()
    }

    gunicorn_server = StandaloneApplication(options, skip_build_assets)
    gunicorn_server.run()


if __name__ == '__main__':
    main()



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from gevent import monkey, sleep, spawn
monkey.patch_all()  # NOQA

from sqlalchemy import create_engine
from collections import deque
import logging
import logging.handlers
import ujson
import errno
import time
import os

from iris.api import load_config
from iris import metrics

# metrics
stats_reset = {
    'sql_errors': 0,
    'deleted_messages': 0,
    'deleted_incidents': 0,
    'deleted_comments': 0
}

# logging
logger = logging.getLogger()
formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s %(message)s')
log_file = os.environ.get('RETENTION_LOG_FILE')
if log_file:
    ch = logging.handlers.RotatingFileHandler(log_file, mode='a', maxBytes=10485760, backupCount=10)
else:
    ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logger.setLevel(logging.INFO)
logger.addHandler(ch)


# pidfile
pidfile = os.environ.get('RETENTION_PIDFILE')
if pidfile:
    try:
        pid = os.getpid()
        with open(pidfile, 'w') as h:
            h.write('%s\n' % pid)
            logger.info('Wrote pid %s to %s', pid, pidfile)
    except IOError:
        logger.exception('Failed writing pid to %s', pidfile)

# Avoid using DictCursor; manually handle columns/offsets here, and only create dict
# when time to archive and dump json. XXX: make sure ID is first
incident_fields = (
    ('`incident`.`id`', 'incident_id'),
    ('`incident`.`created`', 'created'),
    ('`incident`.`context`', 'context'),
    ('`incident`.`plan_id`', 'plan_id'),
    ('`plan`.`name`', 'plan_name'),
    ('`application`.`name`', 'application_name'),
    ('`target`.`name`', 'owner'),
)

message_fields = (
    ('`message`.`id`', 'message_id'),
    ('`message`.`incident_id`', 'incident_id'),
    ('`mode`.`name`', 'mode'),
    ('`priority`.`name`', 'priority'),
    ('`target`.`name`', 'target'),
    ('`template`.`name`', 'template'),
    ('`message`.`subject`', 'subject'),
    ('`message`.`template_id`', 'template_id'),
    ('`message`.`body`', 'body'),
    ('`message`.`created`', 'created'),
)

comment_fields = (
    ('`comment`.`id`', 'comment_id'),
    ('`comment`.`incident_id`', 'incident_id'),
    ('`target`.`name`', 'author'),
    ('`comment`.`content`', 'content'),
    ('`comment`.`created`', 'created'),
)


def archive_incident(incident_row, archive_path):
    incident = {field[1]: incident_row[i] for i, field in enumerate(incident_fields)}

    created = incident['created']
    incident_dir = os.path.join(archive_path, str(created.year), str(created.month), str(created.day), str(incident['incident_id']))

    try:
        os.makedirs(incident_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            logger.exception('Failed creating %s DIR', incident_dir)
            return

    incident_file = os.path.join(incident_dir, 'incident_data.json')

    try:
        with open(incident_file, 'w') as handle:
            ujson.dump(incident, handle, indent=2)
    except IOError:
        logger.exception('Failed writing incident metadata to %s', incident_file)


def archive_message(message_row, archive_path):
    message = {field[1]: message_row[i] for i, field in enumerate(message_fields)}

    created = message['created']
    incident_dir = os.path.join(archive_path, str(created.year), str(created.month), str(created.day), str(message['incident_id']))

    try:
        os.makedirs(incident_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            logger.exception('Failed creating %s DIR', incident_dir)
            return

    message_file = os.path.join(incident_dir, 'message_%d.json' % message['message_id'])

    try:
        with open(message_file, 'w') as handle:
            ujson.dump(message, handle, indent=2)
    except IOError:
        logger.exception('Failed writing message to %s', message_file)


def archive_comment(comment_row, archive_path):
    comment = {field[1]: comment_row[i] for i, field in enumerate(comment_fields)}

    created = comment['created']
    incident_dir = os.path.join(archive_path, str(created.year), str(created.month), str(created.day), str(comment['incident_id']))

    try:
        os.makedirs(incident_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            logger.exception('Failed creating %s DIR', incident_dir)
            return

    comment_file = os.path.join(incident_dir, 'comment_%d.json' % comment['comment_id'])

    try:
        with open(comment_file, 'w') as handle:
            ujson.dump(comment, handle, indent=2)
    except IOError:
        logger.exception('Failed writing comment to %s', comment_file)


def process_retention(engine, max_days, batch_size, cooldown_time, archive_path):
    time_start = time.time()

    connection = engine.raw_connection()
    cursor = connection.cursor(engine.dialect.dbapi.cursors.SSCursor)

    deleted_incidents = 0
    deleted_messages = 0
    deleted_comments = 0

    # First, archive/kill incidents and their messages
    while True:

        # Get incidents to archive and kill, in batches
        try:
            cursor.execute(
                '''
                    SELECT
                        %s
                    FROM `incident`
                    LEFT JOIN `plan` on `plan`.`id` = `incident`.`plan_id`
                    LEFT JOIN `application` on `application`.`id` = `incident`.`application_id`
                    LEFT JOIN `target` ON `incident`.`owner_id` = `target`.`id`
                    WHERE `incident`.`created` < (CURDATE() - INTERVAL %%s DAY)
                    LIMIT %%s
                ''' % (', '.join(field[0] for field in incident_fields)),
                [max_days, batch_size])
        except Exception:
            logger.exception('Failed getting incidents')
            try:
                cursor.close()
            except Exception:
                pass
            cursor = connection.cursor(engine.dialect.dbapi.cursors.SSCursor)
            break

        incident_ids = deque()

        for incident in cursor:
            archive_incident(incident, archive_path)
            incident_ids.append(incident[0])

        if not incident_ids:
            break

        logger.info('Archived %d incidents', len(incident_ids))

        # Then, Archive+Kill all comments in these incidents
        while True:

            try:
                cursor.execute(
                    '''
                        SELECT
                          %s
                        FROM `comment`
                        LEFT JOIN `target` ON `comment`.`user_id` = `target`.`id`
                        WHERE `comment`.`incident_id` in %%s
                        LIMIT %%s
                    ''' % (', '.join(field[0] for field in comment_fields)),
                    [tuple(incident_ids), batch_size])

            except Exception:
                metrics.incr('sql_errors')
                logger.exception('Failed getting comments')
                try:
                    cursor.close()
                except Exception:
                    pass
                cursor = connection.cursor(engine.dialect.dbapi.cursors.SSCursor)
                break

            comment_ids = deque()

            for comment in cursor:
                archive_comment(comment, archive_path)
                comment_ids.append(comment[0])

            if not comment_ids:
                break

            logger.info('Archived %d comments', len(comment_ids))

            try:
                deleted_rows = cursor.execute('DELETE FROM `comment` WHERE `id` IN %s', [tuple(comment_ids)])
                connection.commit()
            except Exception:
                metrics.incr('sql_errors')
                logger.exception('Failed deleting comments from incidents')
                try:
                    cursor.close()
                except Exception:
                    pass
                cursor = connection.cursor(engine.dialect.dbapi.cursors.SSCursor)
                break
            else:
                if deleted_rows:
                    logger.info('Killed %d comments from %d incidents', deleted_rows, len(incident_ids))
                    deleted_comments += deleted_rows
                    sleep(cooldown_time)
                else:
                    break

        # Kill all dynamic plan maps associated with these incidents
        while True:
            try:
                deleted_rows = cursor.execute('DELETE FROM `dynamic_plan_map` WHERE `incident_id` IN %s', [tuple(incident_ids)])
                connection.commit()
            except Exception:
                metrics.incr('sql_errors')
                logger.exception('Failed deleting dynamic plan maps')
                try:
                    cursor.close()
                except Exception:
                    pass
                cursor = connection.cursor(engine.dialect.dbapi.cursors.SSCursor)
                break
            else:
                if deleted_rows:
                    logger.info('Killed %d dynamic plan maps', deleted_rows)
                    deleted_messages += deleted_rows
                    sleep(cooldown_time)
                else:
                    break

        # Archive+Kill all messages in these incidents
        while True:

            try:
                cursor.execute(
                    '''
                        SELECT
                          %s
                        FROM `message`
                        JOIN `priority` on `priority`.`id` = `message`.`priority_id`
                        LEFT JOIN `mode` on `mode`.`id` = `message`.`mode_id`
                        LEFT JOIN `template` ON `message`.`template_id` = `template`.`id`
                        LEFT JOIN `target` ON `message`.`target_id` = `target`.`id`
                        WHERE `message`.`incident_id` in %%s
                        LIMIT %%s
                    ''' % (', '.join(field[0] for field in message_fields)),
                    [tuple(incident_ids), batch_size])

            except Exception:
                metrics.incr('sql_errors')
                logger.exception('Failed getting messages')
                try:
                    cursor.close()
                except Exception:
                    pass
                cursor = connection.cursor(engine.dialect.dbapi.cursors.SSCursor)
                break

            message_ids = deque()

            for message in cursor:
                archive_message(message, archive_path)
                message_ids.append(message[0])

            if not message_ids:
                break

            logger.info('Archived %d messages', len(message_ids))

            # explicitly delete all the extra message data
            try:
                cursor.execute('DELETE FROM `message_changelog` WHERE `message_id` IN %s', [tuple(message_ids)])
                cursor.execute('DELETE FROM `response` WHERE `message_id` IN %s', [tuple(message_ids)])
                cursor.execute('DELETE FROM `twilio_delivery_status` WHERE `message_id` IN %s', [tuple(message_ids)])
                cursor.execute('DELETE FROM `twilio_retry` WHERE `message_id` IN %s', [tuple(message_ids)])
                cursor.execute('DELETE FROM `generic_message_sent_status` WHERE `message_id` IN %s', [tuple(message_ids)])
                connection.commit()
            except Exception:
                metrics.incr('sql_errors')
                logger.exception('Failed deleting message child')
                try:
                    cursor.close()
                except Exception:
                    pass
                cursor = connection.cursor(engine.dialect.dbapi.cursors.SSCursor)

            try:
                deleted_rows = cursor.execute('DELETE FROM `message` WHERE `id` IN %s', [tuple(message_ids)])
                connection.commit()
            except Exception:
                metrics.incr('sql_errors')
                logger.exception('Failed deleting messages from incidents')
                try:
                    cursor.close()
                except Exception:
                    pass
                cursor = connection.cursor(engine.dialect.dbapi.cursors.SSCursor)
                # try deleting individually to directly identify any issues and prevent single error from stopping cleanup
                deleted_rows = 0
                for msg_id in message_ids:
                    try:
                        deleted_rows += cursor.execute('DELETE FROM `message` WHERE `id`=%s', msg_id)
                        connection.commit()
                    except Exception:
                        metrics.incr('sql_errors')
                        logger.exception('Failed deleting message id: %s', msg_id)
            else:
                if deleted_rows:
                    logger.info('Killed %d messages from %d incidents', deleted_rows, len(incident_ids))
                    deleted_messages += deleted_rows
                    sleep(cooldown_time)
                else:
                    break

        # Finally kill incidents
        try:
            deleted_rows = cursor.execute('DELETE FROM `incident` WHERE `id` IN %s', [tuple(incident_ids)])
            connection.commit()
        except Exception:
            metrics.incr('sql_errors')
            logger.exception('Failed deleting incidents')
            try:
                cursor.close()
            except Exception:
                pass
            cursor = connection.cursor(engine.dialect.dbapi.cursors.SSCursor)
            # try deleting individually to directly identify any issues and prevent single error from stopping clean-up
            deleted_rows = 0
            for inc_id in incident_ids:
                try:
                    deleted_rows += cursor.execute('DELETE FROM `incident` WHERE `id`=%s', inc_id)
                    connection.commit()
                except Exception:
                    metrics.incr('sql_errors')
                    logger.exception('Failed deleting incident id: %s', inc_id)
        logger.info('Deleted %s incidents', deleted_rows)
        deleted_incidents += deleted_rows
        sleep(cooldown_time)

    # Next, kill messages not tied to incidents, like quota notifs or incident tracking emails
    while True:
        try:
            deleted_rows = cursor.execute('DELETE FROM `message` WHERE `created` < (CURDATE() - INTERVAL %s DAY) AND `incident_id` IS NULL LIMIT %s', [max_days, batch_size])
            connection.commit()
        except Exception:
            metrics.incr('sql_errors')
            logger.exception('Failed deleting messages')
            try:
                cursor.close()
            except Exception:
                pass
            break
        else:
            if deleted_rows:
                logger.info('Killed %d misc messages', deleted_rows)
                deleted_messages += deleted_rows
                sleep(cooldown_time)
            else:
                break

    try:
        cursor.close()
    except Exception:
        pass
    connection.close()

    logger.info('Run took %.2f seconds and deleted %d incidents and %d messages', time.time() - time_start, deleted_incidents, deleted_messages)
    metrics.set('deleted_messages', deleted_messages)
    metrics.set('deleted_incidents', deleted_incidents)
    metrics.set('deleted_comments', deleted_comments)


def main():
    config = load_config()
    metrics.init(config, 'iris-process-retention', stats_reset)

    retention_settings = config.get('retention', {})
    if not retention_settings.get('enabled'):
        logger.info('Retention not enabled, bailing')
        return

    if config.get('db_retention'):
        engine = create_engine(config['db_retention']['conn']['str'] % config['db_retention']['conn']['kwargs'],
                               **config['db_retention']['kwargs'])
    else:
        engine = create_engine(config['db']['conn']['str'] % config['db']['conn']['kwargs'],
                               **config['db']['kwargs'])

    max_days = int(retention_settings['max_days'])
    if max_days < 1:
        logger.error('Max days needs to at least be 1')
        return

    cooldown_time = float(retention_settings['cooldown_time'])
    batch_size = int(retention_settings['batch_size'])
    run_interval = int(retention_settings['run_interval'])
    archive_path = retention_settings['archive_path']

    spawn(metrics.emit_forever)

    while True:
        logger.info('Starting retention loop (kill messages+incidents older than %d days)', max_days)
        try:
            process_retention(engine, max_days=max_days, cooldown_time=cooldown_time, batch_size=batch_size, archive_path=archive_path)
        except Exception:
            logger.exception('Hit problem while running retention')
        logger.info('Waiting %d seconds until next iteration..', run_interval)
        sleep(run_interval)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.


def test_twilio_message_generate(mocker):
    from iris.vendors.iris_twilio import iris_twilio
    twilio = iris_twilio({})
    assert twilio.generate_message_text({}) == ''
    assert twilio.generate_message_text({'subject': '', 'body': ''}) == ''
    assert twilio.generate_message_text({'subject': 'Foo', 'body': 'Bar'}) == 'Foo. Bar'
    assert twilio.generate_message_text({'body': 'FooBar'}) == 'FooBar'
    assert twilio.generate_message_text({'subject': 'FooBar'}) == 'FooBar'


def test_twilio_notification_call_generate(mocker):
    mocker.patch('iris.vendors.iris_twilio.find_plugin')
    from iris.vendors.iris_twilio import iris_twilio
    relay_base_url = 'http://foo/relay'
    fake_from_num = '123-123-1234'
    fake_to_num = '123-123-1235'

    twilio = iris_twilio({
        'twilio_number': fake_from_num,
        'relay_base_url': relay_base_url,
    })
    mock_twilio_client = mocker.MagicMock()
    mocker.patch.object(twilio, 'get_twilio_client').return_value = mock_twilio_client
    twilio.send_call({
        'destination': fake_to_num,
        'application': 'iris-sender',
        'subject': 'Hello',
        'body': 'World',
    })
    mock_twilio_client.calls.create.assert_called_once_with(
        to=fake_to_num,
        from_=fake_from_num,
        if_machine='Continue',
        url=relay_base_url + (
            '/api/v0/twilio/calls/say?content=Hello.+World&'
            'loop=3&source=iris-sender'),
        status_callback=relay_base_url + '/api/v0/twilio/status'
    )


def test_twilio_incident_call_generate(mocker):
    mock_plugin = mocker.MagicMock()
    mock_plugin.get_phone_menu_text.return_value = 'Press 1 to pay'
    mocker.patch('iris.vendors.iris_twilio.find_plugin').return_value = mock_plugin
    relay_base_url = 'http://foo/relay'
    fake_from_num = '123-123-1234'
    fake_to_num = '123-123-1235'

    from iris.vendors.iris_twilio import iris_twilio
    twilio = iris_twilio({
        'twilio_number': fake_from_num,
        'relay_base_url': relay_base_url,
    })
    mock_twilio_client = mocker.MagicMock()
    mocker.patch.object(twilio, 'get_twilio_client').return_value = mock_twilio_client
    mocker.patch.object(twilio, 'initialize_twilio_message_status')

    twilio.send_call({
        'message_id': 1,
        'destination': fake_to_num,
        'application': 'iris-sender',
        'subject': 'Hello',
        'body': 'World',
    })
    mock_twilio_client.calls.create.assert_called_once_with(
        to=fake_to_num,
        from_=fake_from_num,
        if_machine='Continue',
        url=relay_base_url + (
            '/api/v0/twilio/calls/gather?content=Hello.+World&'
            'loop=3&source=iris-sender&'
            'message_id=1&instruction=Press+1+to+pay'),
        status_callback=relay_base_url + '/api/v0/twilio/status'
    )



# -*- coding:utf-8 -*-
# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.


def test_load_config(mocker):
    import os
    from tempfile import NamedTemporaryFile
    from iris.config import load_config

    mocker.patch.dict(os.environ, {'IRIS_CFG_DB_USER': 'iris_dev'})

    with NamedTemporaryFile() as temp_config:
        temp_config.write(b'''
db:
  conn:
    kwargs:
      scheme: mysql+pymysql
      user: iris
      password: iris
      host: 127.0.0.1
      database: iris
      charset: utf8
    str: "%(scheme)s://%(user)s:%(password)s@%(host)s/%(database)s?charset=%(charset)s"
''')
        temp_config.flush()
        config = load_config(temp_config.name)
        assert config['db']['conn']['kwargs']['user'] == 'iris_dev'



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import logging
from pyfcm import FCMNotification
from iris import db

logger = logging.getLogger(__name__)


class fcm(object):
    def __init__(self, config):
        self.config = config
        self.api_key = self.config.get('api_key')
        # FCM guarantees best-effort delivery with TTL 0
        self.ttl = self.config.get('ttl', 0)
        self.timeout = self.config.get('timeout', 10)
        self.default_notification = self.config.get('notification_title')
        self.proxy = None
        if 'proxy' in self.config:
            host = self.config['proxy']['host']
            port = self.config['proxy']['port']
            self.proxy = {'http': 'http://%s:%s' % (host, port),
                          'https': 'https://%s:%s' % (host, port)}
        self.client = FCMNotification(api_key=self.api_key, proxy_dict=self.proxy)

    def send_push(self, message):
        # Tracking message have no target, skip sending push notification
        if 'target' not in message:
            return
        connection = db.engine.raw_connection()
        cursor = connection.cursor()
        cursor.execute('''SELECT `registration_id`, `platform`
                          FROM `device` WHERE `user_id` =
                          (SELECT `id` FROM `target` WHERE `name` = %s
                          AND `type_id` = (SELECT `id` FROM `target_type` WHERE `name` = 'user'))''',
                       message['target'])
        registration_ids = cursor.fetchall()
        android_ids = [row[0] for row in registration_ids if row[1] == 'Android']
        ios_ids = [row[0] for row in registration_ids if row[1] == 'iOS']
        invalid_ids = []
        failed_ids = []
        # Handle iOS and Android ids separately. Mobile app requires different formats for
        # correct behavior on both platforms (esp with respect to action buttons)
        if ios_ids:
            try:
                data_message = {'incident_id': message.get('incident_id')}
                response = self.client.notify_multiple_devices(
                    registration_ids=ios_ids,
                    message_title=message.get('subject', self.default_notification),
                    message_body=message.get('body', ''),
                    sound='default',
                    time_to_live=self.ttl,
                    data_message=data_message,
                    timeout=self.timeout,
                    click_action='incident'
                )
                for idx, result in enumerate(response['results']):
                    error = result.get('error')
                    if error == 'NotRegistered':
                        invalid_ids.append(ios_ids[idx])
                    elif error is not None:
                        failed_ids.append((ios_ids[idx], error))
            except Exception:
                logger.exception('FCM request failed for message id %s', message.get('message_id'))
        if android_ids:
            try:
                data_message = {'incident_id': message.get('incident_id'),
                                'title': message.get('subject', self.default_notification),
                                'message': message.get('body', ''),
                                'actions': [{
                                    'title': 'Claim',
                                    'callback': 'claimIncident',
                                    'foreground': True
                                }]
                                }
                response = self.client.multiple_devices_data_message(
                    registration_ids=android_ids,
                    time_to_live=self.ttl,
                    data_message=data_message,
                    timeout=self.timeout
                )
                for idx, result in enumerate(response['results']):
                    error = result.get('error')
                    if error == 'NotRegistered':
                        invalid_ids.append(android_ids[idx])
                    elif error is not None:
                        failed_ids.append((android_ids[idx], error))
            except Exception:
                logger.exception('FCM request failed for message id %s', message.get('message_id'))
        # Clean invalidated push notification IDs
        if invalid_ids:
            cursor.execute('''DELETE FROM `device` WHERE `registration_id` IN %s''', (invalid_ids,))
            connection.commit()
        if failed_ids:
            logger.exception('FCM requests failed: %s', failed_ids)
        cursor.close()
        connection.close()



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import logging
import ldap
import os

logger = logging.getLogger(__name__)


class Authenticator:
    def __init__(self, config):
        root = os.path.abspath('./')
        if config.get('debug'):
            self.authenticate = self.debug_auth
            return
        self.authenticate = self.ldap_auth

        if 'ldap_cert_path' in config['auth']:
            self.cert_path = os.path.join(root, config['auth']['ldap_cert_path'])
            if not os.access(self.cert_path, os.R_OK):
                logger.error("Failed to read ldap_cert_path certificate")
                raise IOError
        else:
            self.cert_path = None

        self.bind_user = config['auth'].get('ldap_bind_user')
        self.bind_password = config['auth'].get('ldap_bind_password')
        self.search_filter = config['auth'].get('ldap_search_filter')

        self.ldap_url = config['auth'].get('ldap_url')
        self.base_dn = config['auth'].get('ldap_base_dn')

        self.user_suffix = config['auth'].get('ldap_user_suffix')

    def ldap_auth(self, username, password):
        if self.cert_path:
            ldap.set_option(ldap.OPT_X_TLS_CACERTFILE, self.cert_path)
        connection = ldap.initialize(self.ldap_url)
        connection.set_option(ldap.OPT_REFERRALS, 0)

        if not password:
            return False

        auth_user = username + self.user_suffix
        try:
            if self.bind_user:
                # use search filter to find DN of username
                connection.simple_bind_s(self.bind_user, self.bind_password)
                sfilter = self.search_filter % username
                result = connection.search_s(self.base_dn, ldap.SCOPE_SUBTREE, sfilter, ['dn'])
                if len(result) < 1:
                    return False
                auth_user = result[0][0]

            connection.simple_bind_s(auth_user, password)

        except ldap.INVALID_CREDENTIALS:
            return False
        except (ldap.SERVER_DOWN, ldap.INVALID_DN_SYNTAX) as err:
            logger.warning("%s", err)
            return None
        return True

    def debug_auth(self, username, password):
        return True



#!/usr/bin/env python
# -*- coding:utf-8 -*-


def test_smtp_vendor_init():
    from iris.vendors.iris_smtp import iris_smtp
    smtp_config = {
        'smtp_server': 'foo'
    }
    smtp_vendor = iris_smtp(smtp_config)
    assert smtp_vendor.smtp_timeout == 10
    assert smtp_vendor.mx_sorted == [(0, 'foo')]

    smtp_config = {
        'timeout': 20,
        'smtp_server': 'foo'
    }
    smtp_vendor = iris_smtp(smtp_config)
    assert smtp_vendor.smtp_timeout == 20


def test_smtp_send_email(mocker):
    from iris.vendors.iris_smtp import iris_smtp
    smtp_config = {
        'smtp_server': 'foo',
        'from': 'iris@bar',
    }
    smtp_vendor = iris_smtp(smtp_config)
    mocked_SMPT = mocker.patch('iris.vendors.iris_smtp.SMTP')
    smtp_vendor.send_email({
        'destination': 'foo@bar',
        'subject': 'hello',
        'body': 'world',
    })
    mocked_SMPT.assert_called_once_with(timeout=10)

    class CheckEmailString(str):
        def __eq__(self, other):
            return 'hello' in other and 'world' in other

    mocked_SMPT.return_value.sendmail.assert_called_once_with(
        ['iris@bar'],
        ['foo@bar'],
        CheckEmailString())
    smtp_vendor.cleanup()
    mocked_SMPT.return_value.quit.assert_called_once_with()


def test_smtp_unicode(mocker):
    from iris.vendors.iris_smtp import iris_smtp
    smtp_config = {
        'smtp_server': 'foo',
        'from': 'iris@bar',
    }
    smtp_vendor = iris_smtp(smtp_config)
    mocker.patch('iris.vendors.iris_smtp.SMTP')
    smtp_vendor.send_email({
        'destination': 'foo@bar',
        'subject': 'hello',
        'body': '\u201c',
    })



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

# This is named 'influx' to avoid conflicting with the influxdb module

from datetime import datetime
import logging
from influxdb import InfluxDBClient
from influxdb.exceptions import InfluxDBClientError, InfluxDBServerError
from requests.exceptions import RequestException

logger = logging.getLogger()


# pip install influxdb==3.0.0
class influx(object):
    def __init__(self, config, appname):
        try:
            self.client = InfluxDBClient(**config['influxdb']['connect'])
            self.enable_metrics = True
        except KeyError:
            logger.warning('Missing connect arguments for influxdb. Running with no metrics.')
            self.enable_metrics = False
            return
        try:
            self.extra_tags = config['influxdb']['tags']
        except KeyError:
            self.extra_tags = {}
        self.appname = appname

    def send_metrics(self, metrics):
        if not self.enable_metrics:
            return
        now = str(datetime.now())
        payload = []
        for metric, value in metrics.items():
            data = {
                'measurement': self.appname,
                'tags': {},
                'time': now,
                'fields': {
                    metric: value
                }
            }
            if self.extra_tags:
                data['tags'].update(self.extra_tags)
            payload.append(data)

        try:
            self.client.write_points(payload)
        except (RequestException, InfluxDBClientError, InfluxDBServerError):
            logger.exception('Failed to send metrics to influxdb')



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from docutils import nodes
from docutils.statemachine import ViewList

from sphinx.util import force_decode
from sphinx.util.compat import Directive
from sphinx.util.nodes import nested_parse_with_titles
from sphinx.util.docstrings import prepare_docstring
from sphinx.pycode import ModuleAnalyzer

from sphinxcontrib import httpdomain
from sphinxcontrib.autohttp.common import (
    import_object as autohttp_import_object,
    http_directive as autohttp_http_directive
)


def get_routes(app):
    # deep first tree walk on routing tree
    walk_queue = [node for node in app._router._roots]
    while walk_queue:
        curr_node = walk_queue.pop(0)
        if curr_node.method_map:
            for method in curr_node.method_map:
                handler = curr_node.method_map[method]
                yield method, curr_node.uri_template, handler

        if curr_node.children:
            walk_queue = [chl_node for chl_node in curr_node.children] + walk_queue


class AutofalconDirective(Directive):
    has_content = True
    required_arguments = 1

    def make_rst(self, section_title_set):
        # print('importing falcon app %s...' % self.arguments[0])
        app = autohttp_import_object(self.arguments[0])
        for method, path, handler in get_routes(app):
            docstring = handler.__doc__
            if not isinstance(docstring, str):
                analyzer = ModuleAnalyzer.for_module(handler.__module__)
                docstring = force_decode(docstring, analyzer.encoding)
            if not docstring and 'include-empty-docstring' not in self.options:
                continue
            # exclude falcon HTTPMethodNotAllowed endpoints
            if docstring == 'Raise 405 HTTPMethodNotAllowed error':
                continue
            if not docstring:
                continue
            if hasattr(handler, '__self__'):
                if not (handler.__self__.allow_read_no_auth and method == 'GET'):
                    docstring += '\n:reqheader Authorization: see :ref:`hmac-auth-label`.\n'

            docstring = prepare_docstring(docstring)

            # generate section title if needed
            if path.startswith('/v'):
                section_title = '/'.join(path.split('/')[0:3])
            else:
                section_title = path
            if section_title not in section_title_set:
                section_title_set.add(section_title)
                yield section_title
                yield '_' * len(section_title)

            for line in autohttp_http_directive(method, path, docstring):
                yield line

    def run(self):
        node = nodes.section()
        node.document = self.state.document
        result = ViewList()
        section_title_set = set()
        for line in self.make_rst(section_title_set):
            result.append(line, '<autofalcon>')
        nested_parse_with_titles(self.state, result, node)
        return node.children


def setup(app):
    if 'http' not in app.domains:
        httpdomain.setup(app)
    app.add_directive('autofalcon', AutofalconDirective)



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import logging
from iris.constants import EMAIL_SUPPORT, IM_SUPPORT, CALL_SUPPORT, SMS_SUPPORT, SLACK_SUPPORT

logger = logging.getLogger(__name__)


class iris_dummy(object):
    supports = frozenset([EMAIL_SUPPORT, IM_SUPPORT, CALL_SUPPORT, SMS_SUPPORT, SLACK_SUPPORT])

    def __init__(self, config):
        self.time_taken = 1

    def send(self, message, customizations=None):
        if isinstance(customizations, dict):
            time_taken = customizations.get('time_taken', self.time_taken)
        else:
            time_taken = self.time_taken
        if 'email_subject' in message:
            logger.info('SEND: %(destination)s %(email_subject)s', message)
        else:
            message['application'] = message.get('application')
            message['subject'] = message.get('subject', 'Tracking message')
            logger.info('SEND: %(mode)s %(application)s %(destination)s %(subject).25s', message)
        return time_taken



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import logging
logger = logging.getLogger(__name__)


class user(object):
    def __init__(self, config):
        pass

    def get(self, role, target):
        if role == 'user':
            return [target]
        else:
            return None



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from webassets import script
import sys

from iris.ui import assets_env


def main():
    script.main(sys.argv[1:], env=assets_env)


if __name__ == '__main__':
    main()



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

import requests
import logging
from iris.metrics import stats
import oncallclient
from iris.role_lookup import IrisRoleLookupException

logger = logging.getLogger(__name__)


class oncall(object):
    def __init__(self, config):
        headers = requests.utils.default_headers()
        headers['User-Agent'] = 'iris role lookup (%s)' % headers.get('User-Agent')
        app = config.get('oncall-app', '')
        key = config.get('oncall-key', '')
        self.requests = requests.session()
        if app and key:
            self.requests.auth = oncallclient.OncallAuth(app, key)
        oncallclient.OncallAuth(app, key)
        self.requests.headers = headers
        self.endpoint = config['oncall-api'] + '/api/v0'

    def call_oncall(self, url, expected_type):
        url = str(self.endpoint + url)
        try:
            r = self.requests.get(url)
        except Exception:
            stats['oncall_error'] += 1
            msg = 'Failed hitting oncall-api for url "%s"' % url
            logger.exception(msg)
            raise IrisRoleLookupException(msg)

        if r.status_code != 200:
            # 422 from Oncall indicates nonexistent team from a misconfigured plan; return None for this case
            if r.status_code == 422:
                logger.warning('422 for url "%s", likely invalid plan' % url)
                return None
            else:
                stats['oncall_error'] += 1
                msg = 'Invalid response from oncall-api for URL "%s". Code: %s. Content: "%s"' \
                    % (url, r.status_code, r.content)
                logger.error(msg)
                raise IrisRoleLookupException(msg)

        try:
            data = r.json()
            if not isinstance(data, expected_type):
                stats['oncall_error'] += 1
                msg = 'Invalid data recieved from oncall-api for URL %s' % url
                logger.error(msg)
                raise IrisRoleLookupException(msg)
            return data
        except ValueError:
            stats['oncall_error'] += 1
            msg = 'Failed decoding json from oncall-api. URL: "%s" Code: %s' % (url, r.status_code)
            logger.exception(msg)
            raise IrisRoleLookupException(msg)

    def get(self, role, target):
        if role == 'team':
            result = self.call_oncall('/teams/%s/rosters' % target, dict)
            user_list = []
            if result:
                for roster in result:
                    for user in result[roster]['users']:
                        user_list.append(user['name'])
                return user_list
            else:
                return None
        elif role == 'manager':
            result = self.call_oncall('/teams/%s/oncall/manager' % target, list)
            if result:
                return [user['user'] for user in result]
            else:
                return None
        # "oncall" role maps to primary, otherwise grab shift type from the role suffix (eg "oncall-shadow")
        elif role.startswith('oncall'):
            oncall_type = 'primary' if role == 'oncall' else role[7:]
            result = self.call_oncall('/teams/%s/oncall/%s' % (target, oncall_type), list)
            if result:
                return [user['user'] for user in result]
            else:
                return None
        # Other roles can't be handled by this plugin; return None.
        else:
            return None



#!/usr/bin/env python
# -*- coding:utf-8 -*-

from iris.vendors.iris_messagebird import iris_messagebird


def test_message_construction_for_incident():
    messagebird_vendor = iris_messagebird({
        'access_key': 'abc',
    })
    fake_msg = {
        'application': 'grafana',
        'incident_id': 123,
        'body': 'test body',
        'message_id': 456,
        'destination': '0612342341'
    }
    fake_msg['mode'] = 'call'
    msg_payload = messagebird_vendor.get_message_payload(fake_msg)

    assert msg_payload['originator'] == 'Iris'
    assert msg_payload['recipients'] == '%s' % fake_msg['destination']
    assert msg_payload['body'] == '%s' % fake_msg['body']
    assert 'repeat' in msg_payload

    fake_msg['mode'] = 'sms'
    msg_payload = messagebird_vendor.get_message_payload(fake_msg)
    assert 'repeat' not in msg_payload



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.

from time import time
from gevent import spawn, sleep
from gevent.lock import Semaphore
from collections import deque
from datetime import datetime
import iris.cache
from iris import metrics
import logging
import ujson

logger = logging.getLogger(__name__)

get_application_quotas_query = '''SELECT `application`.`name` as application,
                                         `application_quota`.`hard_quota_threshold`,
                                         `application_quota`.`soft_quota_threshold`,
                                         `application_quota`.`hard_quota_duration`,
                                         `application_quota`.`soft_quota_duration`,
                                         `target`.`name` as target_name,
                                         `target_type`.`name` as target_role,
                                         `application_quota`.`plan_name`,
                                         `application_quota`.`wait_time`
                                 FROM `application_quota`
                                 JOIN `application` ON `application`.`id` = `application_quota`.`application_id`
                                 LEFT JOIN `target` on `target`.`id` = `application_quota`.`target_id`
                                 LEFT JOIN `target_type` on `target_type`.`id` = `target`.`type_id` '''

insert_application_quota_query = '''INSERT INTO `application_quota` (`application_id`, `hard_quota_threshold`,
                                                                     `soft_quota_threshold`, `hard_quota_duration`,
                                                                     `soft_quota_duration`, `plan_name`,
                                                                     `target_id`, `wait_time`)
                                    VALUES (:application_id, :hard_quota_threshold, :soft_quota_threshold,
                                            :hard_quota_duration, :soft_quota_duration, :plan_name, :target_id, :wait_time)
                                    ON DUPLICATE KEY UPDATE `hard_quota_threshold` = :hard_quota_threshold,
                                                            `soft_quota_threshold` = :soft_quota_threshold,
                                                            `hard_quota_duration` = :hard_quota_duration,
                                                            `soft_quota_duration` = :soft_quota_duration,
                                                            `plan_name` = :plan_name,
                                                            `target_id` = :target_id,
                                                            `wait_time` = :wait_time'''

create_incident_query = '''INSERT INTO `incident` (`plan_id`, `created`, `context`, `current_step`, `active`, `application_id`)
                           VALUES ((SELECT `plan_id` FROM `plan_active` WHERE `name` = :plan_name),
                                   :created, :context, 0, TRUE, :sender_app_id)'''

check_incident_claimed_query = '''SELECT `active` FROM `incident` WHERE `id` = :id'''

required_quota_keys = frozenset(['hard_quota_threshold', 'soft_quota_threshold',
                                 'hard_quota_duration', 'soft_quota_duration',
                                 'plan_name', 'wait_time', 'target_name'])

quota_int_keys = ('hard_quota_threshold', 'soft_quota_threshold',
                  'hard_quota_duration', 'soft_quota_duration', 'wait_time')

soft_quota_notification_interval = 1800


class ApplicationQuota(object):

    def __init__(self, db, expand_targets, message_send_enqueue, sender_app):
        self.db = db
        self.expand_targets = expand_targets
        self.message_send_enqueue = message_send_enqueue
        self.iris_application = None
        if sender_app:
            self.iris_application = iris.cache.applications.get(sender_app)
            if self.iris_application:
                logger.info('Using iris application (%s) for sender quota notifications.', sender_app)
            else:
                logger.error('Invalid iris application (%s) used for sender. Quota breach notificiations/incidents will not work.', sender_app)
        else:
            logger.warning('Iris sender_app not configured so notifications for quota breaches will not work')

        self.rates = {}  # application: (hard_buckets, soft_buckets, hard_limit, soft_limit, wait_time, plan_name, (target_name, target_role))
        self.last_incidents = {}  # application: (incident_id, time())
        self.last_incidents_mutex = Semaphore()
        self.last_soft_quota_notification_time = {}  # application: time()
        self.last_soft_quota_notification_time_mutex = Semaphore()
        metrics.add_new_metrics({'quota_hard_exceed_cnt': 0, 'quota_soft_exceed_cnt': 0})
        spawn(self.refresh)

    def get_new_rules(self):
        session = self.db.Session()
        for row in session.execute(get_application_quotas_query):
            yield row
        session.close()

    def refresh(self):
        while True:
            logger.info('Refreshing app quotas')
            new_rates = {}

            for application, hard_limit, soft_limit, hard_duration, soft_duration, target_name, target_role, plan_name, wait_time in self.get_new_rules():
                new_rates[application] = (hard_limit, soft_limit, hard_duration // 60, soft_duration // 60, wait_time, plan_name, (target_name, target_role))

            old_keys = self.rates.keys()
            new_keys = new_rates.keys()

            # Remove old application entries
            for key in old_keys - new_keys:
                logger.info('Pruning old application quota for %s', key)
                try:
                    del(self.rates[key])
                    del(self.last_incidents[key])
                except KeyError:
                    pass

            # Create new ones with fresh buckets
            for key in new_keys - old_keys:
                hard_limit, soft_limit, hard_duration, soft_duration, wait_time, plan_name, target = new_rates[key]
                self.rates[key] = (deque([0] * hard_duration, maxlen=hard_duration),  # hard buckets
                                   deque([0] * soft_duration, maxlen=soft_duration),  # soft buckets
                                   hard_limit, soft_limit, wait_time, plan_name, target)

            # Update existing ones + append new time interval. Keep same time bucket object if duration hasn't changed, otherwise create new
            # one and resize accordingly
            for key in new_keys & old_keys:
                hard_limit, soft_limit, hard_duration, soft_duration, wait_time, plan_name, target = new_rates[key]
                self.rates[key] = (self.rates[key][0] if len(self.rates[key][0]) == hard_duration else deque(self.rates[key][0], maxlen=hard_duration),
                                   self.rates[key][1] if len(self.rates[key][1]) == soft_duration else deque(self.rates[key][1], maxlen=soft_duration),
                                   hard_limit, soft_limit, wait_time, plan_name, target)

                # Increase minute interval for hard + soft buckets
                self.rates[key][0].append(0)
                self.rates[key][1].append(0)

            metrics.add_new_metrics({'app_%s_quota_%s_usage_pct' % (app, quota_type): 0 for quota_type in ('hard', 'soft') for app in new_keys})

            logger.info('Refreshed app quotas: %s', ', '.join(new_keys))
            sleep(60)

    def allow_send(self, message):
        application = message.get('application')

        if not application:
            return True

        # Purpose of quotas is to protect downstreams. If we're already going to drop this message,
        # don't let it account against quota.
        if message.get('mode') == 'drop':
            return True

        rate = self.rates.get(application)

        if not rate:
            return True

        hard_buckets, soft_buckets, hard_limit, soft_limit, wait_time, plan_name, target = rate

        # Increment both buckets for this minute
        hard_buckets[-1] += 1
        soft_buckets[-1] += 1

        # If hard limit breached, disallow sending this message and create incident
        hard_quota_usage = sum(hard_buckets)

        hard_usage_pct = 0
        if hard_limit > 0:
            hard_usage_pct = (hard_quota_usage // hard_limit) * 100
        metrics.set('app_%s_quota_hard_usage_pct' % application, hard_usage_pct)

        if hard_quota_usage > hard_limit:
            metrics.incr('quota_hard_exceed_cnt')
            with self.last_incidents_mutex:
                self.notify_incident(application, hard_limit, len(hard_buckets), plan_name, wait_time)
            return False

        # If soft limit breached, just notify owner and still send
        soft_quota_usage = sum(soft_buckets)

        soft_usage_pct = 0
        if soft_limit > 0:
            soft_usage_pct = (soft_quota_usage // soft_limit) * 100
        metrics.set('app_%s_quota_soft_usage_pct' % application, soft_usage_pct)

        if soft_quota_usage > soft_limit:
            metrics.incr('quota_soft_exceed_cnt')
            with self.last_soft_quota_notification_time_mutex:
                self.notify_target(application, soft_limit, len(soft_buckets), *target)
            return True

        return True

    def notify_incident(self, application, limit, duration, plan_name, wait_time):
        if not self.iris_application:
            logger.warning('Application %s breached hard quota. Cannot notify owners as application is not set', application)
            return

        if not plan_name:
            logger.error('Application %s breached hard quota. Cannot create iris incident as plan is not set (may have been deleted).', application)
            return

        logger.warning('Application %s breached hard quota. Will create incident using plan %s', application, plan_name)

        session = self.db.Session()

        # Avoid creating new incident if we have an incident that's either not claimed or claimed and wait_time hasn't been exceeded
        last_incident = self.last_incidents.get(application)
        if last_incident:
            last_incident_id, last_incident_created = last_incident
            claimed = session.execute(check_incident_claimed_query, {'id': last_incident_id}).scalar()

            if claimed:
                logger.info('Skipping creating incident for application %s as existing incident %s is not claimed', application, last_incident_id)
                session.close()
                return

            if wait_time and (time() - last_incident_created) < wait_time:
                logger.info('Skipping creating incident for application %s as it is not yet %s seconds since existing incident %s was claimed',
                            application, wait_time, last_incident_id)
                session.close()
                return

        # Make a new incident
        incident_data = {
            'plan_name': plan_name,
            'created': datetime.utcnow(),
            'sender_app_id': self.iris_application['id'],
            'context': ujson.dumps({
                'quota_breach': {
                    'application': application,
                    'limit': limit,
                    'duration': duration
                }
            })
        }

        incident_id = session.execute(create_incident_query, incident_data).lastrowid

        session.commit()
        session.close()

        self.last_incidents[application] = incident_id, time()
        logger.info('Created incident %s', incident_id)

    def notify_target(self, application, limit, duration, target_name, target_role):
        if not self.iris_application:
            logger.warning('Application %s breached soft quota. Cannot notify owners as application is not set', application)
            return

        if not target_name or not target_role:
            logger.error('Application %s breached soft quota. Cannot notify owner as they aren\'t set (may have been deleted).', application)
            return

        last_notification_time = self.last_soft_quota_notification_time.get(application)

        now = time()
        if last_notification_time is not None and (now - last_notification_time) < soft_quota_notification_interval:
            return

        self.last_soft_quota_notification_time[application] = now

        logger.warning('Application %s breached soft quota. Will notify %s:%s', application, target_role, target_name)

        targets = self.expand_targets(target_role, target_name)

        if not targets:
            logger.error('Failed resolving %s:%s to notify soft quota breach.', target_role, target_name)
            return

        mode_id = iris.cache.modes.get('email')

        if not mode_id:
            logger.error('Failed resolving email mode to notify soft quota breach for application %s', application)
            return

        for username in targets:
            message = {
                'application': self.iris_application['name'],
                'mode_id': mode_id,
                'mode': 'email',
                'target': username,
                'subject': 'Application %s exceeding message quota' % application,
                'body': ('Hi %s\n\nYour application %s is currently exceeding its soft quota of %s messages per %s minutes.\n\n'
                         'If this continues, your messages will eventually be dropped on the floor and an Iris incident will be raised.\n\n'
                         'Regards,\nIris') % (username, application, limit, duration, )
            }
            self.message_send_enqueue(message)
